/**
	$Id: MAPlotFactory.java 6319 2013-09-09 07:56:59Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions.plot;

import java.util.ArrayList;
import java.util.List;

import net.sf.basedb.clients.web.extensions.AbstractJspActionFactory;
import net.sf.basedb.clients.web.servlet.PlotServlet;
import net.sf.basedb.clients.web.util.HTML;
import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.extensions.InvokationContext;
import net.sf.basedb.util.extensions.xml.PathSetter;
import net.sf.basedb.util.extensions.xml.VariableSetter;

/**
	Factory for generating MA and correction factor plots for 
	2-channel data. This factory will only generate plots if
	the current bioassay set has data in the database.
	<p>
	If the conditions are fulfilled the factory generates MA plots
	for each bioassay in the bioassay set. The image generation
	is handled by the standard plot servlet: {@link PlotServlet}.
	<p>
	If the bioassay set is not a root bioassayset the factory
	also generates correction factor plots.
	
	@author Nicklas
	@version 2.12
	@base.modified $Date: 2013-09-09 09:56:59 +0200 (Mon, 09 Sep 2013) $
*/
public class MAPlotFactory
	extends AbstractJspActionFactory<OverviewPlotAction>
{

	private String servletPath = "../plotter/plot";
	private int width = 400;
	private int height = 300;

	public MAPlotFactory()
	{}

	/*
		From the ActionFactory interface
		--------------------------------
	*/
	/**
		Checks if the bioassay set has 2-channel data in the database.
	 */
	@Override
	public boolean prepareContext(InvokationContext<? super OverviewPlotAction> context) 
	{
		BioAssaySet bas = (BioAssaySet)context.getClientContext().getCurrentItem();
		if (bas.getNumSpots() <= 0) return false;
		if (bas.getRawDataType().getChannels() != 2) return false;
		return super.prepareContext(context);
	}

	@Override
	public OverviewPlotAction[] getActions(
			InvokationContext<? super OverviewPlotAction> context)
	{
		// Get context information
		String ID = context.getClientContext().getSessionControl().getId();
		DbControl dc = context.getClientContext().getDbControl();
		BioAssaySet source = (BioAssaySet)context.getClientContext().getCurrentItem();
		BioAssaySet parent = source.getTransformation().getSource();
		boolean hasCfPlot = parent != null && parent.getDataCubeNo() == source.getDataCubeNo();
		ItemQuery<BioAssay> baQuery = source.getBioAssays();
		baQuery.order(Orders.asc(Hql.property("name")));
		List<BioAssay> bioAssays = baQuery.list(dc);
		
		// Plot generators for each bioassay
		PlotGenerator[] maGenerators = new PlotGenerator[bioAssays.size()];
		List<PlotGenerator> cfGenerators = hasCfPlot ? new ArrayList<PlotGenerator>(bioAssays.size()) : null;
		
		// Generate plot url:s for each bioassay
		String xLabel = HTML.urlEncode("A, log10(ch1 * ch2) / 2");
		String yLabel = HTML.urlEncode("M, log2(ch1 / ch2)");
		String M = HTML.urlEncode("log2(ch(1) / ch(2))");
		String A = HTML.urlEncode("log(ch(1) * ch(2)) / 2");
		String maFilter = HTML.urlEncode("ch(1) > 0 && ch(2) > 0");
		int index = 0;
		for (BioAssay ba : bioAssays)
		{
			// MA plot
			PlotGeneratorBean maPlot = new PlotGeneratorBean();
			int id = ba.getId();
			int subdir = id % 1000;
			maPlot.setId("maplot." + id);
			maPlot.setWidth(width);
			maPlot.setHeight(height);
			maPlot.setTitle("MA plot for " + ba.getName());
			maPlot.setUrl(
					servletPath + "?ID=" + ID + 
					"&bioassay_id=" + id + 
					"&type=scatter"+
					"&width="+width+"&height=" + height +
					"&cache=overview/" + subdir + "/maplot-" + id + ".png" +
					"&x="+A+"&y="+M+
					"&xLabel="+xLabel+"&yLabel="+yLabel +
					"&filter="+maFilter +
					"&title="+HTML.urlEncode(ba.getName())
			);
			maGenerators[index] = maPlot;
			// Correction factor plot
			if (hasCfPlot && ba.getParents().count(dc) == 1)
			{
				PlotGeneratorBean cfPlot = new PlotGeneratorBean();
				cfPlot.setId("cfplot." + id);
				cfPlot.setWidth(width);
				cfPlot.setHeight(height);
				cfPlot.setTitle("Correction factor plot for " + ba.getName());
				cfPlot.setUrl(
						servletPath + "?ID=" + ID + 
						"&bioassay_id=" + id + 
						"&type=cfplot"+
						"&width="+width+"&height=" + height +
						"&cache=overview/" + subdir + "/cfplot-" + id + ".png" +
						"&xLabel="+xLabel+"&yLabel="+yLabel +
						"&title="+HTML.urlEncode(ba.getName())
				);
				cfGenerators.add(cfPlot);
			}
			++index;
		}

		// Main plot actions
		hasCfPlot = cfGenerators != null && cfGenerators.size() > 0;
		OverviewPlotAction[] actions = new OverviewPlotAction[hasCfPlot ? 2 : 1];

		// MA plot
		OverviewPlotBean maBean = new OverviewPlotBean();
		maBean.setTitle("MA plots");
		maBean.setDescription("MA plots for each bioassay in this bioassay set");
		maBean.setId("maplots");
		maBean.setPlotGenerators(maGenerators);
		actions[0] = maBean;
		
		if (hasCfPlot)
		{
			OverviewPlotBean cfBean = new OverviewPlotBean();
			cfBean.setTitle("Correction factor plots");
			cfBean.setDescription("MA-plot of the parent bioassays together " +
					"with a correction factor for each spot in the current bioassayset.");
			cfBean.setId("cfplots");
			cfBean.setPlotGenerators(cfGenerators.toArray(new PlotGenerator[cfGenerators.size()]));
			actions[1] = cfBean;
		}
		
		return actions;
	}
	// --------------------------------

	@VariableSetter
	@PathSetter
	public void setPlotServler(String servletPath)
	{
		this.servletPath = servletPath;
	}
	
	public void setWidth(String width)
	{
		this.width = Values.getInt(width, 400);
	}
	
	public void setHeight(String height)
	{
		this.height = Values.getInt(height, 300);
	}
	
}
