/**
	$Id: Services.java 7703 2019-04-11 07:31:26Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions.service;

import net.sf.basedb.core.Application;
import net.sf.basedb.core.ServicesUtil;
import net.sf.basedb.util.extensions.Extension;
import net.sf.basedb.util.extensions.ExtensionsFilter;
import net.sf.basedb.util.extensions.ExtensionsInvoker;
import net.sf.basedb.util.extensions.Registry;
import net.sf.basedb.util.extensions.SingleExtensionFilter;
import net.sf.basedb.util.extensions.events.EventFilter;
import net.sf.basedb.util.extensions.events.EventHandler;
import net.sf.basedb.util.extensions.events.EventType;
import net.sf.basedb.util.extensions.events.ExtensionPointEventFilter;
import net.sf.basedb.util.extensions.manager.ExtensionsManager;

/**
	Utility class for starting and stopping services.

	@author nicklas
	@version 2.8
	@base.modified $Date: 2019-04-11 09:31:26 +0200 (Thu, 11 Apr 2019) $
*/
public class Services
{

	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(Services.class);

	public static final String EXTENSION_POINT_ID = "net.sf.basedb.clients.web.services";
	
	/**
		Events of this type are sent to the extensions registry whenever a
		service extension is started. Extensions that need to take an action
		whenever this event occur should register themselves with
		{@link Registry#registerEventHandler(EventHandler, EventFilter, ClassLoader)}.
		@since 3.10
	*/
	public static final EventType SERVICE_STARTED = new EventType("Service started");

	/**
		Events of this type are sent to the extensions registry whenever a
		service extension is stopped. Extensions that need to take an action
		whenever this event occur should register themselves with
		{@link Registry#registerEventHandler(EventHandler, EventFilter, ClassLoader)}.
		@since 3.10
	*/
	public static final EventType SERVICE_STOPPED = new EventType("Service stopped");

	private static boolean initialised;
	
	/**
		Initialise and start the installed services. Called when the web server
		is started up.
	*/
	public static synchronized void init()
	{
		if (initialised) return;
		log.info("Initialising services");
		ExtensionsManager manager = Application.getExtensionsManager();
		ExtensionsInvoker<ServiceControllerAction> invoker = getInvoker(null);
		invoker.renderWithEvent(new StartServiceRenderer(), SERVICE_STARTED);
		
		manager.getRegistry().registerEventHandler(
				new ServicesEventHandler(), 
				new ExtensionPointEventFilter(EXTENSION_POINT_ID), null);
		log.info("Initialising services done");
		initialised = true;
	}
	
	/**
		Stop all services.
	*/
	public static synchronized void close()
	{
		if (!initialised) return;
		log.info("Closing all services");
		ExtensionsInvoker<ServiceControllerAction> invoker = getInvoker(null);
		invoker.renderWithEvent(new StopServiceRenderer(), SERVICE_STOPPED);
		initialised = false;
		log.info("Closing all services done");
	}

	/**
		Start services as a result of an administrator has installed a new extension
		or updated/enabled an existing on.
		@param ext The extension, or null to start all enabled extension
		@since 3.3 (was protected before that)
	*/
	public static synchronized void start(Extension<ServiceControllerAction> ext)
	{
		if (!initialised) return;
		log.info("Starting services in extension: " + ext);
		ExtensionsInvoker<ServiceControllerAction> invoker = getInvoker(ext);
		invoker.renderWithEvent(new StartServiceRenderer(), SERVICE_STARTED);
		log.info("Starting services in extension done: " + ext);
	}
	
	/**
		Start services as a result of an administrator has uninstalled an extension
		or is updating/disabling an existing on.
		@param ext The extension, or null to stop all enabled extension
		@since 3.3 (was protected before that)
	*/
	public static synchronized void stop(Extension<ServiceControllerAction> ext)
	{
		if (!initialised) return;
		log.info("Stopping services in extension: " + ext);
		ExtensionsInvoker<ServiceControllerAction> invoker = getInvoker(ext);
		invoker.renderWithEvent(new StopServiceRenderer(), SERVICE_STOPPED);
		log.info("Stopping services in extension done: " + ext);
	}
	
	/**
		Restart services as a result of an administrator has uninstalled an extension
		or is updating/disabling an existing on.
		@param ext The extension, or null to stop all enabled extension
		@since 3.3
	*/
	public static synchronized void restart(Extension<ServiceControllerAction> ext)
	{
		if (!initialised) return;
		log.info("Restarting services in extension: " + ext);
		ExtensionsInvoker<ServiceControllerAction> invoker = getInvoker(ext);
		invoker.renderWithEvent(new StopServiceRenderer(), SERVICE_STOPPED);
		invoker.renderWithEvent(new StartServiceRenderer(), SERVICE_STARTED);
		log.info("Restarting services in extension done: " + ext);
	}

	@SuppressWarnings("unchecked")
	private static ExtensionsInvoker<ServiceControllerAction> getInvoker(Extension<ServiceControllerAction> extension)
	{
		ExtensionsFilter filter = extension == null ?
				Application.getExtensionsManager().getSettings() : new SingleExtensionFilter(extension.getId());
		ExtensionsInvoker<?> invoker = ServicesUtil.useExtensions(filter);
		invoker.setClearErrors(true);
		return (ExtensionsInvoker<ServiceControllerAction>)invoker;
	}
	
}
