/*
	$Id$

	Copyright (C) 2020 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.formatter;

import net.sf.basedb.clients.web.Base;
import net.sf.basedb.clients.web.extensions.fileviewer.FileViewerUtil;
import net.sf.basedb.clients.web.taglib.Page;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Nameable;
import net.sf.basedb.util.formatter.Formatter;

/**
	Formatter implementation that generates a HTML output trying to
	link items with best supported method depending on their type.

	* Files are linked with {@link Base#getLinkedFile(String, File, boolean, boolean, boolean, String)}
	  and with a {@link FileViewerUtil} if it is available
	* Nameable items are linked with {@link Base#getLinkedName(String, Nameable, boolean, boolean)}
	* Other items are linked with {@link Base#getLink(String, String, Item, int, boolean)}
	
	
	@author nicklas
	@since 3.17
	@see Base#getLinkedFile(String, File, boolean, boolean, boolean, String)
*/
public class AutoLinkFormatter 
	implements Formatter<BasicItem>
{
	
	private final boolean enableEditLink;
	private final boolean enableViewDownload;
	private final FileViewerUtil fileViewer;
	
	public AutoLinkFormatter() 
	{
		this(true, true);
	}
	
	public AutoLinkFormatter(FileViewerUtil fileViewer) 
	{
		this.fileViewer = fileViewer;
		this.enableEditLink = true;
		this.enableViewDownload = true;
	}

	public AutoLinkFormatter(boolean enableEditLink, boolean enableViewDownload)
	{
		this.enableEditLink = enableEditLink;
		this.enableViewDownload = enableViewDownload;
		this.fileViewer = null;
	}

	@Override
	public String format(BasicItem item) 
	{
		String link = "";
		if (item instanceof File)
		{
			if (fileViewer != null)
			{
				link = Base.getLinkedName(null, (File)item, false, enableEditLink) + " " + fileViewer.renderAsString();
			}
			else
			{
				link = Base.getLinkedFile(null, (File)item, false, enableEditLink, enableViewDownload, Page.ROOT());
			}
		}
		else if (item instanceof Nameable)
		{
			link = Base.getLinkedName(null, (Nameable)item, false, enableEditLink);
		}
		else if (item != null)
		{
			link = Base.getLink(null, item.toString(), item.getType(), item.getId(), enableEditLink);
		}
		return link;
	}

	@Override
	public File parseString(String value) 
	{
		throw new UnsupportedOperationException("parseString: " + value);
	}

}
