/*
	$Id: Button.java 7703 2019-04-11 07:31:26Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib;

import java.util.HashMap;
import java.util.Map;
import java.util.regex.Pattern;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;

import net.sf.basedb.clients.web.util.HTML;

/**
	This tag is used to display a clickable button.
	<p>
	<b>Example:</b>
	<pre>
   &lt;%@ taglib prefix="base" uri="/WEB-INF/base.tld" %&gt;
   &lt;base:button onclick="alert('click')" title="Click here"&gt;
   &lt;/base:button&gt;
</pre>

	<p>
	<b>Syntax:</b>
	<pre>
   &lt;base:button
      id=...
      clazz=...
      subclass=...
      style=...
      title=...
      tooltip=...
      image=...
      visible=true|false
      disabled=true|false
      tabindex=...
   &gt;
</pre>

	<table class="as-grid">
	<caption>Tag attributes</caption>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>id</td>
		<td>-</td>
		<td>no</td>
		<td>The ID of the button. The value of this attribute
		goes directly into the standard <code>id</code> attribute.
		This means that if you need a reference to the button object in your own
		JavaScript code, you may for example use the following code:
		<p>
		<code>
		note = document.getElementById('&lt;id&gt;');
		</code>
		</td>
	</tr>
	<tr>
		<td>clazz</td>
		<td>basicbutton</td>
		<td>no</td>
		<td>
			The value of this attribute goes directly into the standard HTML 
			<code>class</code> attribute. This allows you to apply different styles
			to different buttons.
		</td>
	</tr>
	<tr>
		<td>subclass</td>
		<td>-</td>
		<td>no</td>
		<td>
			Additional classes to put in the HTML class attribute. Doesn't
			replace the main class attribute.
		</td>
	</tr>
	<tr>
		<td>style</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute. This allows you to apply different styles
			to different buttons.
		</td>
	</tr>
	<tr>
		<td>title</td>
		<td></td>
		<td>no</td>
		<td>The title of the button.</td>
	</tr>
	<tr>
		<td>tooltip</td>
		<td></td>
		<td>no</td>
		<td>
			A short explanation that will show up as a tooltip.
		</td>
	</tr>
	<tr>
		<td>image</td>
		<td></td>
		<td>no</td>
		<td>
			An optional image to display. An image is automatically
			selected if the title is "Close", "Cancel", "Save", "Ok", "Yes" or "No".
		</td>
	</tr>
	<tr>
		<td>visible</td>
		<td>true</td>
		<td>no</td>
		<td>
			If the button should be visible or not.
		</td>
	</tr>
	<tr>
		<td>disabled</td>
		<td>false</td>
		<td>no</td>
		<td>
			If the button should be enabled or not.
		</td>
	</tr>
	<tr>
		<td>tabindex</td>
		<td>0</td>
		<td>no</td>
		<td>
			Tab index value to control the focus order. Use -1 to disable tabbing to the button.
		</td>
	</tr>
	</table>

	@author Nicklas
	@version 2.0
*/
public class Button
	extends StylableTag
{

	/**
		
	*/
	private static final long serialVersionUID = 2058163945669701874L;

	private static final Map<String, String> defaultImages = new HashMap<String, String>();
	private static final Map<String, String> defaultIds = new HashMap<String, String>();
	static
	{
		defaultImages.put("cancel", "cancel.png");
		defaultImages.put("close", "close.png");
		defaultImages.put("save", "save.png");
		defaultImages.put("ok", "ok.png");
		defaultImages.put("yes", "yes.png");
		defaultImages.put("no", "no.png");
		defaultImages.put("back", "goback.png");
		defaultImages.put("next", "gonext.png");
		defaultImages.put("add", "add.png");
		defaultImages.put("remove", "remove.png");
		defaultImages.put("browse", "browse.png");
		defaultImages.put("select", "select.png");
		
		defaultIds.put("cancel", "close");
		defaultIds.put("close", "close");
	}
	
	public Button()
	{
		super("basicbutton button interactable");
	}
	
	private ButtonGroup buttonGroup;

	/**
		The title of the button.
	*/
	private String title = null;

	/**
		An optional tooltip.
	*/
	private String tooltip = null;

	/**
		If the button should be visible or not.
	*/
	private boolean visible = true;

	/**
		If the button should be enabled or not.
	*/
	private boolean disabled = false;

	/**
		The image to display.
	*/
	private String image = null;
	
	private int tabIndex = 0;
	
	/*
		Taglib initialization methods
	*/
	@Override
	public String getId() 
	{
		String id = super.getId();
		if (id == null && getTitle() != null)
		{
			id = getDefaultId(getTitle());
		}
		return id;
	}
	
	public void setTitle(String title)
	{
		this.title = title;
	}
	public String getTitle()
	{
		return title;
	}
	
	public void setImage(String image)
	{
		this.image = image;
	}
	public String getImage()
	{
		return image;
	}

	public void setTooltip(String tooltip)
	{
		this.tooltip = tooltip;
	}
	public String getTooltip()
	{
		return tooltip;
	}

	public void setVisible(boolean visible)
	{
		this.visible = visible;
	}
	public boolean isVisible()
	{
		return visible;
	}

	public void setDisabled(boolean disabled)
	{
		this.disabled = disabled;
	}
	public boolean isDisabled()
	{
		return disabled;
	}
	
	/**
		@since 3.2
	*/
	public void setTabindex(int tabIndex)
	{
		this.tabIndex = tabIndex;
	}
	
	/**
		@since 3.2
	*/
	public int getTabindex()
	{
		return tabIndex;
	}

	@Override
	public int doStartTag()
		throws JspException
	{
		if (!isVisible()) return SKIP_BODY;
		buttonGroup = (getParent() instanceof ButtonGroup) ? (ButtonGroup)getParent() : null;
		Page page = (Page)findAncestorWithClass(this, Page.class);

		String theImage = getImage();
		if (theImage == null && getTitle() != null)
		{
			theImage = getDefaultImage(getTitle());
		}
		
		StringBuilder sb = new StringBuilder();
		sb.append("<div");
		addIdAndStyles(sb, isDisabled() ? "disabled" : null);
		if (getId() != null)
		{
			// Assume onclick event is attached by javascript
			if (!isDisabled() && getTabindex() >= 0) 
			{
				sb.append(" tabindex=\"").append(getTabindex()).append("\"");
			}
		}
		if (getTooltip() != null) sb.append(" title=\"").append(getTooltip()).append("\"");
		addDynamicAttributes(sb);
		sb.append(">");
		if (theImage != null)
		{
			if (!theImage.startsWith("/"))
			{
				theImage = page.getRoot() + "images/" + theImage;
			}
			sb.append("<img src=\"").append(theImage).append("\"");
			sb.append(" alt=\"").append(HTML.encodeTags(getTitle())).append("\"");
			sb.append(">");
		}
		if (getTitle() != null)	
		{
			sb.append(getTitle());
		}
		sb.append("</div>");

		try
		{
			if (buttonGroup != null)
			{
				buttonGroup.addButton(sb.toString());
			}
			else
			{
				pageContext.getOut().print(sb.toString());
			}
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return SKIP_BODY;
	}

	private static final Pattern keepFirstWord = Pattern.compile("^(\\w+).*$");
	
	private String getDefaultImage(String theTitle)
	{
		// Convert to lower case and keep only the first word
		theTitle = keepFirstWord.matcher(theTitle).replaceAll("$1").toLowerCase();
		return defaultImages.get(theTitle);
	}

	private String getDefaultId(String theTitle)
	{
		// Convert to lower case and keep only the first word
		theTitle = keepFirstWord.matcher(theTitle).replaceAll("$1").toLowerCase();
		return defaultIds.get(theTitle);
	}


}




