/*
	$Id: Select.java 7703 2019-04-11 07:31:26Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib;

import java.util.Collections;
import java.util.List;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;

import net.sf.basedb.clients.web.util.HTML;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.Directory;
import net.sf.basedb.core.Nameable;
import net.sf.basedb.core.Removable;

/**
	This tag is used to display a selection list of items. The list
	can be linked to a selection popup and has options for the current
	item and recently used items.
	<p>
	<b>Example:</b>
	<pre>
   &lt;%@ taglib prefix="base" uri="/WEB-INF/base.tld" %&gt;
   &lt;base:select 
   	id="protocol"
   	required="false"
   	current="&lt;%=currentProtocol%&gt;"
   	denied="false"
   	recent="&lt;%=recentProtocols%&gt;"
   /&gt;
 </pre>

	<p>
	<b>Syntax:</b>
	<pre>
   &lt;base:select
      id=...
      clazz=...
      style=...
      buttonclass=...
      buttonstyle=...
      buttontitle=...
      buttonicon=...
      required=true|false
      current=...
      denied=true|false
      recent=...
      selectrecent=true|false
      defaultitem=...
      selectdefault=true|false
      newitem=true|false
      visible=true|false
      disabled=true|false
      tabindex=...
   &gt;
</pre>

	<table class="as-grid">
	<caption>Tag attributes</caption>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>id</td>
		<td>-</td>
		<td>yes</td>
		<td>The ID of the selection list. The value of this attribute
		goes directly into the standard <code>name</code> attribute of a &lt;select&gt;
		tag.
		</td>
	</tr>
	<tr>
		<td>clazz</td>
		<td></td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>class</code> attribute. This allows you to apply different styles
			to different selection lists.
		</td>
	</tr>
	<tr>
		<td>style</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute. This allows you to apply different styles
			to different selection lists.
		</td>
	</tr>
	<tr>
		<td>buttonclass</td>
		<td>buttonclass</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>class</code> attribute for the "Select" button part. This allows you to
			apply different styles to the button. If the button is disabled "_disabled" is
			appended to the class name.
		</td>
	</tr>
	<tr>
		<td>buttonstyle</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute for the "Select" button part. This allows you to
			apply different styles to the button. 
		</td>
	</tr>
	<tr>
		<td>buttontitle</td>
		<td>Select</td>
		<td>no</td>
		<td>
			The text on the "Select" button.
		</td>
	</tr>
	<tr>
		<td>buttonicon</td>
		<td>select.png</td>
		<td>no</td>
		<td>
			The icon to use on the "Select" button.
		</td>
	</tr>
	<tr>
		<td>required</td>
		<td>false</td>
		<td>no</td>
		<td>
			If a value must be selected or not. If FALSE is specified
			the option <code>- none -</code> will be included.
		</td>
	</tr>
	<tr>
		<td>current</td>
		<td>null</td>
		<td>yes</td>
		<td>The current item that is selected. It should be a {@link BasicItem}.</td>
	</tr>
	<tr>
		<td>denied</td>
		<td>false</td>
		<td>no</td>
		<td>
			If null is passed as the current item, this parameter should be TRUE
			if the reason is that the logged in user is denied read access, FALSE
			otherwise.
		</td>
	</tr>
	<tr>
		<td>recent</td>
		<td></td>
		<td>no</td>
		<td>
			A list of recently used items. The should be of the same type as
			the current item.
		</td>
	</tr>
	<tr>
		<td>selectrecent</td>
		<td>true</td>
		<td>no</td>
		<td>
			If no current has been specified and it is a new item then, if this
			parameter is TRUE the first recently used value is selected. If this
			parameter is FALSE, no item is selected.
		</td>
	</tr>
	<tr>
		<td>defaultitem</td>
		<td></td>
		<td>no</td>
		<td>
			Default value set e.g. for a project.
		</td>
	</tr>
	<tr>
		<td>selectdefault</td>
		<td>true</td>
		<td>no</td>
		<td>
			If no current or recent has been specified and it is a new item then, if this
			parameter is TRUE the default value is selected. If this
			parameter is FALSE, no item is selected.
		</td>
	</tr>
	<tr>
		<td>newitem</td>
		<td>false</td>
		<td>no</td>
		<td>
			If the parent item is a new item or an already saved item. This
			affects the ID of the current item that is used in the list. For
			existing values the ID is negative, which means that the link
			doesn't have to change unless another item is selected.
		</td>
	</tr>
	<tr>
		<td>visible</td>
		<td>true</td>
		<td>no</td>
		<td>
			If the selection list should be visible or not.
		</td>
	</tr>
	<tr>
		<td>disabled</td>
		<td>false</td>
		<td>no</td>
		<td>
			If the selection list should be enabled or not.
		</td>
	</tr>
	<tr>
		<td>tabindex</td>
		<td>0</td>
		<td>no</td>
		<td>
			Tab index value to control the focus order. Use -1 to disable tabbing to the control.
		</td>
	</tr>
	</table>

	@author Nicklas
	@version 2.0
*/
public class Select
	extends StylableTag
{

	
	/**
		
	*/
	private static final long serialVersionUID = -9096279950931870965L;

	private String buttonclass = "basicbutton button";
	private String buttonstyle = null;
	private String buttontitle = "Select&hellip;";
	private String buttonicon = "select.png";
	
	/**
		If a value is required or not.
	*/
	private boolean required;
	
	/**
		The currently selected item.
	*/
	private transient BasicItem current;
	
	/**
		If the logged in user was denied access to the current item.
	*/
	private boolean denied = false;
	
	/**
		List of recently used items.
	*/
	private transient List<? extends BasicItem> recent;
	
	/**
		If we can select the most recently used item for new items
		with nu current item.
	*/
	private boolean selectRecent = true;
	
	/**
	  	The project default values.
	*/
	private transient List<? extends BasicItem> defaultItems;
	
	/**
	 	If the default item should be selected for new items with
	 	no current or recently used items.
	*/
	private boolean selectdefault = true;
	
	/**
		If the link is from a new item or an old item.
	*/
	private boolean newItem = false;
	
	/**
		If the select list should be visible or not.
	*/
	private boolean visible = true;

	/**
		If the selection list should be enabled or not.
	*/
	private boolean disabled = false;

	/**
		Text to display that for the "unselected" option
	*/
	private String unselectedText = "- none -";
	
	private int tabIndex;
	
	private boolean tryGetDirectoryPath;
	
	public Select()
	{
		super("selectionlist");
	}
	
	/*
		Taglib initialization methods
	*/
	/**
		@since 3.0
	*/
	public void setButtonclass(String clazz)
	{
		this.buttonclass = clazz;
	}
	/**
		@since 3.0
	*/
	public String getButtonclass()
	{
		return buttonclass;
	}
	/**
		@since 3.0
	*/
	public void setButtonstyle(String style)
	{
		this.buttonstyle = style;
	}
	/**
		@since 3.0
	*/
	public String getButtonstyle()
	{
		return buttonstyle;
	}
	/**
		@since 3.0
	*/
	public void setButtontitle(String title)
	{
		this.buttontitle = title;
	}
	/**
		@since 3.0
	*/
	public String getButtontitle()
	{
		return buttontitle;
	}
	/**
		@since 3.0
	*/
	public void setButtonicon(String icon)
	{
		this.buttonicon = icon;
	}
	/**
		@since 3.0
	*/
	public String getButtonicon()
	{
		return buttonicon;
	}
	
	public void setRequired(boolean required)
	{
		this.required = required;
	}
	public boolean isRequired()
	{
		return required;
	}

	public void setCurrent(BasicItem current)
	{
		this.current = current;
	}
	public BasicItem getCurrent()
	{
		return current;
	}
	
	public void setDenied(boolean denied)
	{
		this.denied = denied;
	}
	public boolean isDenied()
	{
		return denied;
	}

	public void setRecent(List<? extends BasicItem> recent)
	{
		this.recent = recent;
	}
	public List<? extends BasicItem> getRecent()
	{
		return recent;
	}
	
	public void setSelectrecent(boolean selectRecent)
	{
		this.selectRecent = selectRecent;
	}
	public boolean getSelectrecent()
	{
		return selectRecent;
	}

	public void setDefaultitem(BasicItem defaultitem)
	{
		this.defaultItems = defaultitem == null ? null : Collections.singletonList(defaultitem);
	}
	
	public void setDefaultitems(List<? extends BasicItem> defaultItems)
	{
		this.defaultItems = defaultItems;
	}
	
	public void setSelectdefault(boolean selectdefault)
	{
		this.selectdefault = selectdefault;
	}
	public boolean getSelectdefault()
	{
		return selectdefault;
	}
	
	public void setNewitem(boolean newItem)
	{
		this.newItem = newItem;
	}
	public boolean isNewitem()
	{
		return newItem;
	}
	
	public void setVisible(boolean visible)
	{
		this.visible = visible;
	}
	public boolean isVisible()
	{
		return visible;
	}

	public void setDisabled(boolean disabled)
	{
		this.disabled = disabled;
	}
	public boolean isDisabled()
	{
		return disabled;
	}

	public void setUnselectedtext(String unselectedText)
	{
		this.unselectedText = unselectedText;
	}
	
	public String getUnselectedtext()
	{
		return unselectedText;
	}
	
	/**
		@since 3.2
	*/
	public void setTabindex(int tabIndex)
	{
		this.tabIndex = tabIndex;
	}
	
	public int getTabindex()
	{
		return tabIndex;
	}
	
	@Override
	public int doStartTag()
		throws JspException
	{
		if (!isVisible()) return SKIP_BODY;
		Page page = (Page)findAncestorWithClass(this, Page.class);

		tryGetDirectoryPath = true;
		
		StringBuilder sb = new StringBuilder();

		sb.append("<div");
		addIdAndStyles(sb);
		addDynamicAttributes(sb);
		sb.append(">\n");
		sb.append("<table>\n");
		sb.append("<tr><td>");
		boolean disabled = isDenied() || isDisabled();
		sb.append("<select name=\"").append(getId()).append("\"");
		sb.append(" id=\"").append(getId()).append(".list\"");
		sb.append(disabled ? " disabled" : isRequired() ? " class=\"required\"" : "");
		if (getTabindex() != 0)	sb.append(" tabindex=\"").append(getTabindex()).append("\"");
		sb.append(">\n");
		if (isDenied())
		{
			sb.append("<option value=\"NaN\">- denied -\n");
		}
		else
		{
			if (!isRequired())
			{
				sb.append("<option value=\"0\">").append(getUnselectedtext()).append("\n");
			}
			boolean hasSelected = false;
			if (current != null)
			{
				sb.append("<option value=\"");
				sb.append(current.getId() * (isNewitem() ? 1 : -1));
				sb.append("\" selected>");
				hasSelected = true;
				String name = getName(current);
				sb.append(HTML.encodeTags(name)).append("\n");
			}
			if (recent != null && recent.size() > 0)
			{
				sb.append("<option value=\"NaN\" disabled class=\"recentheader\">› recently used\n");
				int i = 1;
				for (BasicItem item : recent)
				{
					if (item == null) 
					{
						if (!hasSelected && isNewitem() && getSelectrecent()) 
						{
							// This will cause the 'none' option to be selected by being the first element
							hasSelected = true; 
						}
						continue;
					}
					String name = getName(item);
					String clazz = "selectoptionindent recent";
					if (item instanceof Removable && ((Removable)item).isRemoved())
					{
						clazz += " removed";
						name += " (in trashcan)";
					}
					sb.append("<option value=\"").append(item.getId()).append("\"");
					sb.append(" class=\"").append(clazz).append("\"");
					if (!hasSelected && isNewitem() && getSelectrecent()) 
					{
						sb.append(" selected");
						hasSelected = true;
					}
					sb.append(">").append(i).append(". ").append(name).append("\n");
					i++;
				}
			}
			if (defaultItems != null && defaultItems.size() > 0)
			{
				sb.append("<option value=\"NaN\" disabled class=\"defaultheader\">› project default\n");
				for (BasicItem defaultitem : defaultItems)
				{
					String name = getName(defaultitem);
					
					sb.append("<option value=\"").append(defaultitem.getId()).append("\"");
					sb.append(" class=\"default selectoptionindent\"");				
					if (!hasSelected && isNewitem() && getSelectdefault())
					{					
						sb.append(" selected");
						hasSelected = true;
					}
					sb.append(">").append(name).append("\n");
				}
			}
		}
		sb.append("</select>\n");
		sb.append("</td>");
		
		sb.append("<td>");
		sb.append("<div class=\"interactable ").append(getButtonclass()).append(disabled ? " disabled": "").append("\"");
		sb.append(" id=\"").append(getId()).append(".select\"");
		sb.append(" data-list-name=\"").append(getId()).append("\"");
		if (!disabled)
		{
			if (getTabindex() >= 0)
			{
				sb.append(" tabindex=\"").append(getTabindex()).append("\"");
			}
		}
		if (getButtonstyle() != null)
		{
			sb.append(" style=\"").append(getButtonstyle()).append("\"");
		}
		sb.append(">");
		String buttonImg = getButtonicon();
		if (buttonImg != null)
		{
			if (!buttonImg.startsWith("/"))
			{
				buttonImg = page.getRoot() + "images/" + buttonImg;
			}
			sb.append("<img src=\"").append(buttonImg).append("\" alt=\"").append(HTML.encodeTags(getButtontitle())).append("\">");
		}
		sb.append(getButtontitle());
		sb.append("</div>");
		sb.append("</td>\n");
		sb.append("</tr></table></div>");

		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return SKIP_BODY;
	}

	private String getName(BasicItem item)
	{
		String name = null;
		if (item instanceof Nameable)
		{
			if (tryGetDirectoryPath && item instanceof Directory)
			{
				try
				{
					name = ((Directory)item).getPath().toString();
				}
				catch (Exception ex)
				{
					tryGetDirectoryPath = false;
				}
			}
			if (name == null) name = ((Nameable)item).getName();
		}
		else
		{
			name = item.toString();
		}
		return name;
	}
	
}




