/*
 $Id: Quota.java 7595 2019-02-22 07:21:20Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

 BASE is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 3
 of the License, or (at your option) any later version.

 BASE is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.core.data.GroupData;
import net.sf.basedb.core.data.QuotaData;
import net.sf.basedb.core.data.QuotaIndex;
import net.sf.basedb.core.data.UserData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.query.EntityQuery;

/**
	This class is used to represent disk quota settings for users
	and groups. All users must have a quota that can be either
	assigned directly to the user or via a group where the user
	is a member.
	<p>
	Quota can be specified as a total for all types/locations, or
	with individual settings for each type/location.
	<p>
	Quota is handled like this:
	<ol>
	<li>A {@link DiskConsumable} item tells BASE
		about the number of bytes, type of quota and location of the
		item.
	<li>BASE (via the {@link OwnedItem} creates a
		{@link DiskUsage} item to hold that information
		and links it with the <code>Quota</code> for the owner
		of the <code>DiskConsumable</code> item.
	<li>If the quota is exceeded, BASE will throw a {@link QuotaException}.
	</ol>
	
	@author enell
	@version 2.0
	@see QuotaData
	@base.modified $Date: 2019-02-22 08:21:20 +0100 (Fri, 22 Feb 2019) $
*/
public class Quota
	extends BasicItem
	implements Nameable, SystemItem, Removable
{
	/**
		The type of item represented by this class.
		@see Item#QUOTA ITEM.QUOTA
		@see #getType() getType()
	*/
	public static final Item TYPE = Item.QUOTA;

	/**
		The ID for the system default quota, which is assigned to new
		users unless another quota is specified.
	*/
	public static final String DEFAULT = "net.sf.basedb.core.Quota.DEFAULT";

	/**
		The ID for the unrestricted quota, which gives users
		unlimited disk space.
	*/
	public static final String UNRESTRICTED = "net.sf.basedb.core.Quota.UNRESTRICTED";

	/**
		Quota is unlimited.
	*/
	public static final long UNLIMITED = -1;

	/**
		Quota is is not defined.
	*/
	public static final long UNDEFINED = -2;

	/**
		This filter will only return the logged in user's quota
		unless the logged in user has generic read permission.
	*/
	private static final QueryRuntimeFilter RUNTIME_FILTER = new QueryRuntimeFilterImpl();

	/**
		Create a new <code>Quota</code> item.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param total The total size inte bytes of this <code>Quota</code>. 
		@return The new <code>Quota</code> item
		@throws PermissionDeniedException This exception is thrown if
			the logged in user doesn't have {@link Permission#CREATE CREATE}
			permission for this type of items
		@throws BaseException This exception is thrown if there is another
			error
	*/
	public static Quota getNew(DbControl dc, long total)
		throws BaseException
	{
		Quota q = dc.newItem(Quota.class);
		q.setName("New quota");
		q.setQuotaValue(QuotaType.getById(dc, SystemItems.getId(QuotaType.TOTAL)), Location.PRIMARY, total);
		return q;
	}
	
	/**
		Get a <code>Quota</code> item when you know the ID.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>Quota</code> item
		@throws ItemNotFoundException If an item with the specified 
			ID is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static Quota getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Quota q = dc.loadItem(Quota.class, id);
		if (q == null) throw new ItemNotFoundException("Quota[id="+id+"]");
		return q;
	}

	/**
		Get a query configured to retrieve quota. If the logged in user has 
		generic READ permission all items are returned, otherwise only the quota item 
		that is assigned to the logged in user.
		
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<Quota> getQuery()
	{
		return new ItemQuery<Quota>(Quota.class, RUNTIME_FILTER);
	}
	
	/**
		Create a new <code>Quota</code> item.
	
		@param data The data class for this item. 
	*/
	Quota(QuotaData data)
	{
		super(data);
	}

	@Override
	QuotaData getData()
	{
		return (QuotaData)getBasicData();
	}
	
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	@Override
	public String getName()
	{
		return getData().getName();
	}
	@Override
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	@Override
	public String getDescription()
	{
		return getData().getDescription();
	}
	@Override
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	@Override
	public boolean isRemoved()
	{
		return getData().getRemovedBy() != null;
	}
	@Override
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemovedBy(removed ? getSessionControl().getLoggedInUserId() : null);
	}
	@Override
	public User getRemovedBy()
		throws PermissionDeniedException, ItemNotFoundException 
	{
		Integer userId = getData().getRemovedBy();
		return userId == null ? null : User.getById(getDbControl(), userId);
	}
	// -------------------------------------------
	/*
		From the SystemItem interface
		-------------------------------------------
	*/
	@Override
	public String getSystemId()
	{
		return getData().getSystemId();
	}
	@Override
	public boolean isSystemItem()
	{
		return getSystemId() != null;
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Checks if:
		<ul>
		<li>a {@link Group} has been assigned this quota
		<li>a {@link User} has been assigned this quota
		</ul>
		@return <code>true</code> if this item is used, 
			<code>false</code> otherwise
		@throws BaseException If there is another error
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		boolean used = false;
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.query.Query<Long> query = HibernateUtil.getPredefinedQuery(session, 
			"GET_USERS_FOR_QUOTA", Long.class, "count(*)");
		/*
		 SELECT {1}
		 FROM UserData u
		 WHERE u.quota = :quota
		 */
		query.setParameter("quota", getId(), TypeWrapper.H_INTEGER);
		used = HibernateUtil.loadData(query) > 0;
		if (!used)
		{
			query = HibernateUtil.getPredefinedQuery(session, 
				"GET_GROUPS_FOR_QUOTA", Long.class, "count(*)");
			/*
			 SELECT {1}
			 FROM GroupData g
			 WHERE g.quota = :quota
			*/
			query.setParameter("quota", getId(), TypeWrapper.H_INTEGER);
			used = HibernateUtil.loadData(query) > 0;
		}
		return used || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link User}:s assigned to this quota
		<li>{@link Group}:s assigned to this quota
		</ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		
		// Users
		org.hibernate.query.Query<Integer> query = HibernateUtil.getPredefinedQuery(session, 
			"GET_USERS_FOR_QUOTA", Integer.class, "u.id");
			/*
				SELECT {1}
				FROM UserData u
				WHERE u.quota = :quota
			*/
		query.setParameter("quota", getId(), TypeWrapper.H_INTEGER);
		addUsingItems(using, Item.USER, query);
		
		// Groups
		query = HibernateUtil.getPredefinedQuery(session, 
			"GET_GROUPS_FOR_QUOTA", Integer.class, "g.id");
			/*
				SELECT {1}
				FROM GroupData g
				WHERE g.quota = :quota
			*/
		query.setParameter("quota", getId(), TypeWrapper.H_INTEGER);
		addUsingItems(using, Item.GROUP, query);
		
		return using;
	}

	/**
		The logged in user has read permission to the quota
		assigned to him/her, directly or through group quota. 
		If this is a system item, delete and create permissions are denied. 
		
		@throws BaseException If there is an error
	*/
	@Override
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		if (isSystemItem()) 
		{
			denied |= Permission.deny(Permission.DELETE, Permission.CREATE);
		}
		SessionControl sc = getSessionControl();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		UserData user = HibernateUtil.loadData(session, UserData.class, sc.getLoggedInUserId());
		if (getId() == user.getQuota().getId())
		{
			granted |= Permission.grant(Permission.READ);
		}
		else
		{
			GroupData quotaGroup = user.getQuotaGroup();
			if (quotaGroup != null)
			{
				QuotaData groupQuota = quotaGroup.getQuota();
				if (groupQuota != null && groupQuota.getId() == getId())
				{
					granted |= Permission.grant(Permission.READ);
				}
			}
		}
		super.initPermissions(granted, denied);
	}
	// -------------------------------------------

	/**
		Set the quota for the specified type, location and size.
		@param quotaType The {@link QuotaType QuotaType} for this <code>Quota</code>
		@param location The location  for this <code>Quota</code>
		@param bytes The maximum number of bytes that can be used, or
			{@link #UNLIMITED UNLIMITED} if the quota should be unlimited,
			or {@link #UNDEFINED UNDEFINED} if the quota setting should be
			removed
		@throws InvalidDataException If any of the parameters is wrong.
		@throws PermissionDeniedException If the user has no write permissions.
	*/
	public void setQuotaValue(QuotaType quotaType, Location location, long bytes)
		throws InvalidDataException, PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		if (quotaType == null)
		{
			throw new InvalidUseOfNullException("A quota must have a quotatype");
		}
		if (location == null)
		{
			throw new InvalidUseOfNullException("A quota must have a location");
		}
		else if (!location.isQuotaLimited())
		{
			throw new InvalidDataException("Quota for '" + location + "' location cannot be set for "+quotaType.getName());
		}
		if ((bytes < 0) && (bytes != UNLIMITED) && (bytes != UNDEFINED))
		{
			throw new NumberOutOfRangeException("bytes", bytes, 0, Long.MAX_VALUE);
		}
		quotaType.checkPermission(Permission.USE);
		QuotaIndex qi = new QuotaIndex(quotaType.getData(), location.getValue());
		if (bytes == UNDEFINED)
		{
			getData().getQuotaValues().remove(qi);
		}
		else
		{
			QuotaIndex existing = qi.find(getData().getQuotaValues());
			if (existing != null)
			{
				existing.setMaxBytes(bytes);
			}
			else
			{
				qi.setMaxBytes(bytes);
				getData().getQuotaValues().add(qi);
			}
		}
	}

	/**
	 	Get the quota for the specified type and location.
		@param quotaType The @link QuotaType QuotaType}
		@param location The location of the quota
		@return The maximum number of bytes that can be used, or
	 		{@link #UNLIMITED UNLIMITED} if the quota is unlimited or
	 		{@link #UNDEFINED UNDEFINED} if the quota is not defined
	*/
	public long getQuotaValue(QuotaType quotaType, Location location)
	{
		QuotaIndex qi = new QuotaIndex(quotaType.getData(), location.getValue());
		Long bytes = qi.findMaxBytes(getData().getQuotaValues());
		return bytes == null ? Quota.UNDEFINED : bytes;
	}
	
	/**
		A runtime filter implementation that limits a query to only
		return the quota assigned to the logged in user unless the logged in user
		has generic read permission.
	*/
	private static class QueryRuntimeFilterImpl
		implements QueryRuntimeFilter
	{

		@Override
		public void enableFilters(QueryRuntimeFilterManager manager, EntityQuery query, DbControl dc)
		{
			SessionControl sc = dc.getSessionControl();
			if (sc.isDenied(query.getItemPermission(), Item.QUOTA))
			{
				manager.enableFilter("denyAll");
			}
			else if (!sc.hasPermission(query.getItemPermission(), Item.QUOTA))
			{
				if (sc.isLoggedIn() && query.getItemPermission() == Permission.READ)
				{
					org.hibernate.Filter filter = manager.enableFilter("memberOf");
					if (filter != null)
					{
						UserData userData = HibernateUtil.loadData(dc.getHibernateSession(), 
							UserData.class, sc.getLoggedInUserId());
						assert userData != null : "userData == null";
						
						Set<Integer> quotaIds = new HashSet<Integer>();
						// Direct quota
						quotaIds.add(userData.getQuota().getId());
						
						// Group quota
						GroupData quotaGroup = userData.getQuotaGroup();
						if (quotaGroup != null)
						{
							QuotaData groupQuota = quotaGroup.getQuota();
							if (groupQuota != null)
							{
								quotaIds.add(groupQuota.getId());
							}
						}
						
						filter.setParameterList("items", quotaIds);
						filter.setParameter("owner", sc.getLoggedInUserId());
					}
				}
				else
				{
					// Do not load any quota if not logged in
					manager.enableFilter("denyAll");
				}
			}
		}
	}
}
