/**
	$Id: AbstractSignalTransporter.java 6444 2014-04-09 12:21:20Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.signal;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.Collection;
import java.util.HashSet;

/**
	Abstract base class that is useful when implementing signal transporters.

	@author nicklas
	@version 2.6
	@base.modified $Date: 2014-04-09 14:21:20 +0200 (Wed, 09 Apr 2014) $
*/
public abstract class AbstractSignalTransporter
	implements SignalTransporter
{
	/**
		Log signals processing.
	*/
	private static final org.slf4j.Logger logger = 
		org.slf4j.LoggerFactory.getLogger(AbstractSignalTransporter.class);
	
	private String globalSignalId;
	private URI signalHandlerURI;
	private Collection<Signal> signals;
	
	protected AbstractSignalTransporter()
	{}
	
	/*
		From the SignalTransporter interface
		-------------------------------------------
	*/
	/**
		Initialise the transporter. We expect the parameter to be the one
		returned from {@link AbstractSignalReceiver#getGlobalSignalId(SignalHandler)}.
		If the receiver subclass overrided that method, the corresponding transporter
		subclass should override this method.
		@see AbstractSignalReceiver for a description of the format expected
	*/
	@Override
	public void init(String params)
	{
		logger.info("Initializing signal transporter: id=" + params);
		globalSignalId = params;
	}
	
	@Override
	public Collection<Signal> getSupportedSignals()
	{
		if (signals == null)
		{
			signals = new HashSet<Signal>();
			URI uri = getSignalURI();
			String query = uri.getQuery();
			if (query != null)
			{
				String[] signalIds = query.split(",");
				for (String id : signalIds)
				{
					Signal s = Signal.getSignal(id);
					if (s != null) signals.add(s);
				}
			}
		}
		return signals;
	}
	// -------------------------------------------

	/**
		Generate a signal message string for the given signal. The string
		will have the following format:
		<code>signal://handlerId@receiverId/?signal</code>
		@see AbstractSignalReceiver#processSignalMessage(String)
	*/
	protected String generateSignalMessage(Signal signal)
	{
		String message = "signal://" + getHandlerId() + "@" + getReceiverId() + "/?" + signal.getId();
		return message;
	}
	
	/**
		Get the raw ID string that was passed to the {@link #init(String)}
		method.
	*/
	protected String getGlobalSignalId()
	{
		return globalSignalId;
	}
	
	/**
		Get the URI representation of the global signal ID.
		@return An URI
		@throws SignalException If the signal ID is not a valid URI.
	*/
	protected URI getSignalURI()
	{
		if (signalHandlerURI == null)
		{
			try
			{
				signalHandlerURI = new URI(globalSignalId);
			}
			catch (URISyntaxException ex)
			{
				throw new SignalException(globalSignalId, ex);
			}
		}
		return signalHandlerURI;
	}
	
	/**
		Get the receiverId part of the signal URI.
		@return The ID of the receiver
	*/
	protected String getReceiverId()
	{
		URI uri = getSignalURI();
		String receiverId = uri.getHost();
		if (uri.getPort() >= 0) receiverId += ":" + uri.getPort();
		return receiverId;
	}
	
	/**
		Get the local handler ID part of the signal URI.
		@return The local ID of the signal handler
	*/
	protected String getHandlerId()
	{
		return getSignalURI().getUserInfo();
	}
	
	
}
