/**
	$Id: InputStreamCombiner.java 5353 2010-05-27 08:54:05Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import java.io.IOException;
import java.io.InputStream;

/**
	An input stream implementation that reads one or more input streams
	sequentially. When the end of the first stream has been reached, reading
	continues on the next one, and so on until all streams have been read.
	The underlying streams are automatically closed once they have been fully 
	read.

	@author Nicklas
	@since 2.16
	@base.modified $Date: 2010-05-27 10:54:05 +0200 (Thu, 27 May 2010) $
*/
public class InputStreamCombiner
	extends InputStream
{

	private final InputStream[] in;
	private int index;
	
	/**
		Create a new stream combiner using the given streams to read from
		@param in The stream to read from, must be an array with at least
			one stream
	*/
	public InputStreamCombiner(InputStream... in)
	{
		this.in = in;
		if (in == null || in.length == 0) throw new NullPointerException("'in' is null or empty");
		this.index = 0;
	}
	
	/*
		From the InputStream class
		--------------------------
	*/
	@Override
	public int available() 
		throws IOException
	{
		if (index >= in.length) return 0;
		return in[index].available();
	}

	/**
		Closes all underlying streams.
	*/
	@Override
	public void close() 
		throws IOException
	{
		while (index < in.length)
		{
			nextStream();
		}
	}

	@Override
	public int read(byte[] b, int off, int len) 
		throws IOException
	{
		int result = -1;
		while (result == -1 && index < in.length)
		{
			result = in[index].read(b, off, len);
			if (result == -1) nextStream();
		}
		return result;		
	}

	@Override
	public int read(byte[] b) 
		throws IOException
	{
		return read(b, 0, b.length);
	}

	@Override
	public long skip(long n) 
		throws IOException
	{
		if (index >= in.length) return 0;
		long skipped = 0;
		while (read() != -1 && skipped < n)
		{
			++skipped;
		}
		return skipped;
	}

	@Override
	public int read() 
		throws IOException
	{
		int result = -1;
		while (result == -1 && index < in.length)
		{
			result = in[index].read();
			if (result == -1) nextStream();
		}
		return result;
	}
	// -------------------------------------

	private void nextStream()
	{
		FileUtil.close(in[index]);
		++index;
	}
	
}
