/**
	$Id $

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.annotations;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Nameable;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.SimpleProgressReporter;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.core.snapshot.SnapshotManager;

/**
	Utility class for processing multiple items with a 
	{@link InheritAnnotationsManager} in a separated thread 
	with progress reporting.
	<p>
	Initialize this class with a session control, a list
	of {@link Annotatable} items and a list of 
	{@link InheritSpecification}:s, then start a new 
	thread: new Thread(manager).start();
	
	@author nicklas
	@since 3.5
*/
public class RunnableInheritAnnotationsManager 
	implements Runnable
{

	private final SessionControl sc;
	private ProgressReporter progress;
	private final List<InheritSpecification> specifications;
	private final List<Annotatable> items;
	
	/**
		Create a new instance that should use the given session
		control.
	*/
	public RunnableInheritAnnotationsManager(SessionControl sc) 
	{
		this.sc = sc;
		this.specifications = new ArrayList<InheritSpecification>();
		this.items = new ArrayList<Annotatable>();
	}
	
	/**
		Set the progress reporter to use for reporting progress.
	*/
	public void setProgressReporter(ProgressReporter progress)
	{
		this.progress = progress;
	}
	
	/**
		Add a specification with rules for inheriting annotations.
	*/
	public void addSpecification(InheritSpecification specification)
	{
		specifications.add(specification);
	}
	
	/**
		Add an item that should inherit annotations. The item will
		be processed with {@link InheritAnnotationsManager#process(Annotatable)}
		once the thread is stared.
	*/
	public void addItem(Annotatable item)
	{
		items.add(item);
	}
	
	/**
		Add all items for processing.
	*/
	public void addAllItems(Collection<? extends Annotatable> items)
	{
		this.items.addAll(items);
	}
	
	@Override
	public void run() 
	{
		DbControl dc = null;
		String itemName = null;
		try
		{
			dc = sc.newDbControl(":Batch inherit annotations");
			InheritAnnotationsManager manager = new InheritAnnotationsManager(dc, new SnapshotManager(), specifications);
			
			int i = 0;
			int numItems = items.size();
			for (Annotatable item : items)
			{
				ThreadSignalHandler.checkInterrupted();
				// Reload item in current transaction
				item = item.getType().getById(dc, item.getId());

				itemName = item instanceof Nameable ? ((Nameable)item).getName() : item.toString();
				
				int percent = (100 * i) / numItems;
				i++;
				if (progress != null)
				{
					progress.display(percent, "Processing (" + i + " of " + numItems + "): " + itemName);
				}
				manager.process(item);
			}
			if (progress != null) progress.display(99, "Committing transaction. Please wait...");
			dc.commit();
			
			if (progress != null) progress.display(100, "Successfully processed " + numItems + " items");
		}
		catch (RuntimeException ex)
		{
			if (progress != null)
			{
				progress.display(100, "Failed while processing " + itemName + ": " + ex.getMessage());
				if (progress instanceof SimpleProgressReporter)
				{
					((SimpleProgressReporter)progress).setError(ex);
				}
			}
			ex.printStackTrace(System.out);
		}
		finally
		{
			if (dc != null) dc.close();
		}
		
	}

}
