package net.sf.basedb.util.excel;

import java.util.Date;

import org.apache.poi.ss.usermodel.Cell;

/**
	Represents a value and style format that should be written to
	a cell in an excel worksheet.
	
	Note that Excel internally only supports String and Double values.
	To get the intended result for integers, dates, times, etc. a
	format string must be applied.
	
	@author nicklas
	@since 3.15
*/
public abstract class ExcelValue<T> 
{

	// Some static cells
	private static final ExcelValue<Void> EMPTY_CELL = new BlankValue<>();
	private static final ExcelValue<Boolean> TRUE = new BooleanValue(true);	
	private static final ExcelValue<Boolean> FALSE = new BooleanValue(false);
	
	/**
		This creates a blank cell.
	*/
	@SuppressWarnings("unchecked")
	public static final <T> ExcelValue<T> emptyCell()
	{
		return (ExcelValue<T>)EMPTY_CELL;
	}
	
	/**
		Create a string cell value.
	*/
	public static ExcelValue<String> asString(String value)
	{
		if (value == null) return emptyCell();
		return new StringValue(value);
	}
	
	/**
		Create a date cell value. The cell will be formatted according
		to the specified format string which should be given as a Java-compatible
		format string. If no format is specified "yyyy-MM-dd" is used.
	*/
	public static ExcelValue<Date> asDate(Date value, String javaFormat)
	{
		if (value == null) return emptyCell();
		if (javaFormat == null) javaFormat = "yyyy-MM-dd";
		return new DateValue(value, javaFormat);
	}
	
	/**
		Create a timestamp cell value. The cell will be formatted according
		to the specified format string which should be given as a Java-compatible
		format string. If no format is specified "yyyy-MM-dd HH:mm:ss" is used.
		Note! There is really no difference between this metod and the {@link #asDate(Date, String)}
		method except for the default format string.
	*/
	public static ExcelValue<Date> asTimestamp(Date value, String javaFormat)
	{
		if (value == null) return emptyCell();
		if (javaFormat == null) javaFormat = "yyyy-MM-dd HH:mm:ss";
		return new DateValue(value, javaFormat);
	}
	
	/**
		Create a numeric cell value without any decimals. 
		Same as calling {@link #asNumber(Number, int)} with
		0 for the numDecimals parameter.
	*/
	public static final ExcelValue<Number> asInt(Number value)
	{
		if (value == null) return emptyCell();
		return new IntValue(value);
	}

	/**
		Create a numeric cell value and display a given number of
		decimals. The numDecimals should be between 0 and 15,
		or the "General" format will be used.
	*/
	public static final ExcelValue<Number> asNumber(Number value, int numDecimals)
	{
		if (value == null) return emptyCell();
		return new NumericValue(value, numberFormat(numDecimals));
	}
	
	/**
		Create a numeric cell value with a custom numeric format.
	*/
	public static final ExcelValue<Number> asNumber(Number value, String format)
	{
		if (value == null) return emptyCell();
		return new NumericValue(value, format);
	}
	
	/**
		Create a numeric cell value with boolean format.
	*/
	public static final ExcelValue<Boolean> asBoolean(Boolean value)
	{
		if (value == null) return emptyCell();
		return value ? TRUE : FALSE;
	}
	
	/**
		Creates a format string for formatting numbers with the given number
		of decimals (between 0 and 15).
	*/
	public static String numberFormat(int numDecimals)
	{
		if (numDecimals < 0 || numDecimals > 15) return null;
		if (numDecimals == 0) return "0";
		StringBuilder sb = new StringBuilder(2+numDecimals);
		sb.append("0.");
		for (int i = 0; i < numDecimals; i++)
		{
			sb.append("0");
		}
		return sb.toString();
	}
	
	protected final T value;
	protected final String format;
	
	protected ExcelValue(T value, String format)
	{
		this.value = value;
		this.format = format;
	}
	
	/**
		Get the stored value.
	*/
	public T getValue()
	{
		return value;
	}
	
	/**
		Get the format string.
	*/
	public String getFormat()
	{
		return format;
	}
	
	/**
		Write the value to the given cell and apply a format to it
		with the help of the style creator. If no style creator is
		given, the cell is not formatted.
	*/
	public abstract void writeTo(Cell cell, CellStyleCreator styleCreator);
	
	/**
		Represents a blank cell.
	*/
	static class BlankValue<T>
		extends ExcelValue<T>
	{
		BlankValue()
		{
			super(null, null);
		}
		@Override
		public void writeTo(Cell cell, CellStyleCreator styleCreator) 
		{
			cell.setBlank();
			cell.setCellStyle(styleCreator.getBaseStyle());
		}		
	}
	
	/**
		Represents a cell with a string value.
	*/
	static class StringValue
		extends ExcelValue<String>
	{
		StringValue(String value)
		{
			super(value, null);
		}

		@Override
		public void writeTo(Cell cell, CellStyleCreator styleCreator) 
		{
			cell.setCellValue(value);
			cell.setCellStyle(styleCreator.getBaseStyle());
		}
	}
	
	/**
		Represents a cell with a date value. Note that date cells are
		actually cells with a numeric value and a date-like format string.
	*/
	static class DateValue
		extends ExcelValue<Date>
	{
		DateValue(Date value, String format)
		{
			super(value, format);
		}

		@Override
		public void writeTo(Cell cell, CellStyleCreator styleCreator) 
		{
			cell.setCellValue(value);
			if (format != null)
			{
				cell.setCellStyle(styleCreator.getDateCellStyle(format));
			}
		}
	}
	
	/**
		Represents a cell with a numeric value without decimals. 
	*/
	static class IntValue
		extends ExcelValue<Number>
	{
		IntValue(Number value)
		{
			super(value, null);
		}
		
		@Override
		public void writeTo(Cell cell, CellStyleCreator styleCreator) 
		{
			cell.setCellValue(value.doubleValue());
			cell.setCellStyle(styleCreator.getIntCellStyle());
		}		
	}
	
	/**
		Represents a cell with a numeric value. 
	*/
	static class NumericValue
		extends ExcelValue<Number>
	{
		NumericValue(Number value, String format)
		{
			super(value, format);
		}
		
		@Override
		public void writeTo(Cell cell, CellStyleCreator styleCreator) 
		{
			cell.setCellValue(value.doubleValue());
			if (format != null)
			{
				cell.setCellStyle(styleCreator.getCellStyle(format));
			}
		}
	}
	
	/**
		Represents a cell with a numeric value. 
	*/
	static class BooleanValue
		extends ExcelValue<Boolean>
	{
		BooleanValue(boolean value)
		{
			super(value, "BOOLEAN");
		}
		
		@Override
		public void writeTo(Cell cell, CellStyleCreator styleCreator) 
		{
			cell.setCellValue(value ? 1 : 0);
			cell.setCellStyle(styleCreator.getCellStyle(format));
		}
	}

}
