/**
	$Id: BaseFileExporter.java 6875 2015-04-20 09:08:18Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.export.spotdata;


import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DatabaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DynamicResultIterator;
import net.sf.basedb.core.DynamicSpotQuery;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.VirtualColumn;
import net.sf.basedb.core.VirtualTable;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.JoinType;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Selects;
import net.sf.basedb.core.query.SqlResult;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.basefile.BaseFileWriter;
import net.sf.basedb.util.formatter.Formatter;
import net.sf.basedb.util.formatter.IdentifiableFormatter;
import net.sf.basedb.util.formatter.ToStringFormatter;

/**
	A superclass for exporters that exports spot data to BASEfile format.
	This class exposes several other configuration parameters that are 
	specific for BASEfile:s. 

	@author Nicklas
	@version 2.12
	@base.modified $Date: 2015-04-20 11:08:18 +0200 (Mon, 20 Apr 2015) $
*/
public abstract class BaseFileExporter
	extends AbstractBioAssaySetExporter
{

	private Map<String, String> parameters;
	private BaseFileWriter out;
	private boolean autoCloseWriters;
	
	protected BaseFileExporter()
	{
		this.parameters = new LinkedHashMap<String, String>();
		this.autoCloseWriters = true;
	}
	
	/*
		Configuration properties
		------------------------
	*/
	/**
		Set the stream were the exported data should be written.
		It is expected that the given writer is a fresh writer and
		that no data has been written to it yet.
	*/
	public void setWriter(BaseFileWriter out)
	{
		this.out = out;
	}
	
	/**
		If this option is set then all writers are automatically closed
		when all data has been writted to them. This setting is enabled by default.
		@since 3.2
	*/
	public void setAutoCloseWriters(boolean autoClose)
	{
		this.autoCloseWriters = autoClose;
	}
	
	/**
		Add a parameter that is exported in the 'settings' section, which
		is the first section in the BASEfile. There can only be one value
		for a given key. The parameters are written to the file in the 
		order they are registered by this method.
		
		@param key The parameter key (if null this method call is ignored)
		@param value The parameter value (may be null)
	*/
	public void setParameter(String key, String value)
	{
		if (key == null) return;
		parameters.put(key, value);
	}
	
	/**
		Get the current value of a 'settings' parameter.
		@param key The parameter key
		@return The value, or null if no values has been registered for
			the key (or if the registered value is null)
	*/
	public String getParameter(String key)
	{
		return parameters.get(key);
	}
	
	@Override
	public void addSpotField(DynamicField field)
	{
		super.addSpotField(field);
	}
	
	/**
		Add multiple spot fields in one go. A null collection is ignored and
		so are null values in the collection.
		@param fields A collection with the fields that should be added
	*/
	public void addSpotFields(Collection<? extends DynamicField> fields)
	{
		if (fields == null) return;
		for (DynamicField f : fields)
		{
			addSpotField(f);
		}
	}
	
	@Override
	public List<DynamicField> getSpotFields()
	{
		return super.getSpotFields();
	}
	
	@Override
	public void addReporterField(DynamicField field)
	{
		super.addReporterField(field);
	}
	
	/**
		Add multiple reporter fields in one go. A null collection is ignored and
		so are null values in the collection.
		@param fields A collection with the fields that should be added
	*/
	public void addReporterFields(Collection<? extends DynamicField> fields)
	{
		if (fields == null) return;
		for (DynamicField f : fields)
		{
			 addReporterField(f);
		}
	}

	@Override
	public List<DynamicField> getReporterFields()
	{
		return super.getReporterFields();
	}
	
	/**
		Adds an assay field to the exported file. If no fields has been added
		the exporter will automaticall export the assay id and name and all 
		experimental factor values.
		@since 2.16
	*/
	@Override
	public void addAssayField(AssayField field)
	{
		super.addAssayField(field);
	}

	/**
		Add multiple assay fields in one go. A null collection is ignored and
		so are null values in the collection. If no fields has been added
		the exporter will automaticall export the assay id and name and all 
		experimental factor values.
		@param fields A collection with the fields that should be added
		@since 2.16
	*/
	public void addAssayFields(Collection<? extends AssayField> fields)
	{
		if (fields == null) return;
		for (AssayField f : fields)
		{
			 addAssayField(f);
		}
	}

	/**
		@since 2.16
	*/
	@Override
	public List<AssayField> getAssayFields()
	{
		return super.getAssayFields();
	}

	
	@Override
	public void setAverageOnReporter(boolean averageOnReporter)
	{
		super.setAverageOnReporter(averageOnReporter);
	}
	@Override
	public boolean getAverageOnReporter()
	{
		return super.getAverageOnReporter();
	}
	// --------------------------------------
	
	/*
		From the AbstractBioAssaySetExporter class
		------------------------------------------
	*/
	/**
		Prepare the export by adding all experimental
		factors as assay fields if no fields have been specified.
	*/
	@Override
	protected void beginExport()
	{
		if (getAssayFields() == null || getAssayFields().size() == 0)
		{
			addAssayField(ExportableFieldFactory.assayProperty("id", "id", Type.INT, null));
			addAssayField(ExportableFieldFactory.assayProperty("name", "name", Type.STRING, null));
			addExperimentalFactorsAsAssayFields();
		}
	}
	/**
		Export 'settings' and 'assays' sections.
	*/
	@Override
	protected boolean exportGlobalHeader()
	{
		exportSettingsSection();
		exportAssaysSectionHeaders();
		exportAssaysSectionData();
		return true;
	}
	
	@Override
	protected void endExport(RuntimeException e) 
	{
		if (out != null && autoCloseWriters) out.close();
		super.endExport(e);
	}
	// -------------------------------------------	

	/**
		Get the writer that writes the data to the file.
	*/
	protected BaseFileWriter getBaseFileWriter()
	{
		return out;
	}
	
	/**
		Export the settings section which contains all parameter values
		registered with {@link #setParameter(String, String)}. If a parameter
		with key=section exists the value of that parameter is used as the 
		name of this section, otherwise 'settings' is used as name.
		If the 'section' parameter exists, it must be the first
		parameter. It is not allowed in other places and will generate an
		exception.
		<p>
		This method leaves the BASEfile writer in the header part in case the caller 
		wants to add more headers. 
		<p>
		If no parameters has been registered this section is skipped.
		
		@return The number of parameters written, not including the 'section'
			parameter
	*/
	protected int exportSettingsSection()
	{
		BaseFileWriter out = getBaseFileWriter();
		out.baseWriteBom();
		if (parameters.isEmpty()) return 0;

		boolean first = true;
		int numParameters = 0;
		for (Map.Entry<String, String> entry : parameters.entrySet())
		{
			String key = entry.getKey();
			String value = entry.getValue();
			String section = null;
			if (first)
			{
				first = false;
				if ("section".equals(key))
				{
					section = value;
					key = null;
				}
				else
				{
					section = "settings";
				}
			}
			if (section != null) out.baseBeginSection(section);
			if (key != null)
			{
				++numParameters;
				out.basePrintHeader(key, value);
			}
		}
		out.flush();
		return numParameters;
	}


	/**
		Exports headers for the 'assays' section which contains metadata 
		information about the assays in the bioassay set. 
		This method leaves the BASEfile writer in the header part in 
		case the caller wants to add more headers. 
	*/
	protected void exportAssaysSectionHeaders()
	{
		// Get configuration settings
		BaseFileWriter out = getBaseFileWriter();
		List<BioAssay> assays = getBioAssays();
		List<AssayField> assayFields = getAssayFields();
		
		// Put together a list with the assay fields that are exported
		int dataColumns = assayFields.size();
		List<String> columns = new ArrayList<String>(dataColumns);
		List<String> annotationColumns = new ArrayList<String>(dataColumns);
		for (AssayField af : assayFields)
		{
			columns.add(af.getTitle());
			if (af.isAnnotation()) annotationColumns.add(af.getTitle());
		}
		
		// Write the headers
		out.baseBeginSection("assays");
		out.basePrintHeader("columns", Values.getString(columns, "\t", false));
		out.basePrintHeader("annotationColumns", Values.getString(annotationColumns, "\t", false));
		out.basePrintHeader("count", assays.size());
		out.flush();
	}
	
	/**
		Export data for the 'assays' section that contains information
		about the bioassays in the source bioassay set. Each data line
		consists of the 'id', 'name' and annotations for a bioassay.
	*/
	@SuppressWarnings({ "unchecked", "rawtypes" })
	protected void exportAssaysSectionData()
	{
		// Get configuration settings
		DbControl dc = getDbControl();
		BaseFileWriter out = getBaseFileWriter();
		List<BioAssay> assays = getBioAssays();
		List<AssayField> assayFields = getAssayFields();
		int dataColumns = assayFields.size();
		
		if (!out.isWritingData()) out.baseBeginDataPart();
		for (BioAssay ba : assays)
		{
			Object[] data = new Object[dataColumns];
			int index = 0;
			for (AssayField at : assayFields)
			{
				Collection<?> values = at.getAssayValue(dc, ba);
				Formatter f = at.getFormatter();
				data[index++] = Values.getString(values, ",", true, f);
			}
			out.basePrintData(data);
		}
		out.flush();
	}

	/**
		Start a new 'spot' section and export the standard headers:
		'channels, 'assays', 'columns', 'assayFields' and 'count'.
		This method leaves the BASEfile writer in the header part in
		case the caller wants to add more headers.
		
		@param assays The bioassays that this section contains data for
		@param spotCount The number of data lines that is going to be
			written in the data part
	*/
	protected void exportSpotSectionHeaders(List<BioAssay> assays, long spotCount)
	{
		// Get configuration settings
		BioAssaySet source = getSource();
		List<DynamicField> reporterFields = getReporterFields();
		List<DynamicField> spotFields = getSpotFields();
		BaseFileWriter out = getBaseFileWriter();
		
		// Write section 'spots' and headers
		out.baseBeginSection("spots");
		out.basePrintHeader("channels", source.getRawDataType().getChannels());
		out.basePrintHeader("assays", Values.getString(assays, "\t", false, new IdentifiableFormatter("")));
		ExportableFieldFormatter f = new ExportableFieldFormatter();
		out.basePrintHeader("columns", Values.getString(reporterFields, "\t", false, f) + "\tassayData");
		out.basePrintHeader("assayFields", Values.getString(spotFields, "\t", false, f));
		out.basePrintHeader("count", spotCount);
	}
	
	/**
		Export the spot data of a 'spot' section. The given query should
		return the given number of spot fields PLUS the spot position and 
		column (in that order). Reporter data is expected to have been cached
		by {@link #cacheReporterData()} and is copied to the output by
		{@link #copyReporterFields(int, Object[], int)}.
		<p>
		A new data row is generated each time the position number changes. If each 
		row should contain data from more than one assay an index map must be given.
		The index map should map the bioassay column to the ordinal number of the
		assay in the list of assays.
		
		@param assays A list of with the bioassays that the current spot
			data section contains data for
		@param spotQuery The query used to retreive the data
		@param progress An optional progress reporter, if given progress
			is reported from 0 to 100%
		@param count The total number of spots must be given for the progress
			reporting to work
	*/
	@SuppressWarnings({ "unchecked", "rawtypes" })
	protected void exportSpotSectionData(List<BioAssay> assays, DynamicSpotQuery spotQuery, 
		ProgressReporter progress, long count)
	{
		// Get configuration options
		DbControl dc = getDbControl();
		BaseFileWriter out = getBaseFileWriter();
		List<DynamicField> spotFields = getSpotFields();
		int numSpotFields = spotFields.size();
		int numReporterFields = getReporterFields().size();
		
		// Prepare temporary storage
		int numAssaysPerRow = assays.size();
		int numColumnsPerRow = numReporterFields + numAssaysPerRow * numSpotFields;
		Object[] data = new Object[numColumnsPerRow];
		// Map assay column number to index in the 'data' array
		if (numAssaysPerRow > 1)
		{
			prepareAssayIndexMap(assays, numReporterFields, numSpotFields);
		}
		Formatter[] formatters = new Formatter[numSpotFields+1];
		for (int i = 0; i < numSpotFields; ++i)
		{
			Formatter f = spotFields.get(i).getFormatter();
			if (f == null) f = new ToStringFormatter();
			formatters[i+1] = f; // Note +1 to avoid re-calc of index later
		}

		if (!out.isWritingData()) out.baseBeginDataPart();
		int posIndex = numSpotFields + 1;
		int colIndex = numSpotFields + 2;
		int currentPosition = -1;
		
		// Execute the query
		if (progress != null) progress.display(0, "Loading spot data...");
		DynamicResultIterator it = spotQuery.iterate(dc);
		try
		{
			long numDone = 0;
			long progressInterval = Math.max(1 + count / 100, 50);
			while (it.hasNext())
			{
				ThreadSignalHandler.checkInterrupted();
				SqlResult result = it.next();
				
				int position = result.getInt(posIndex);
				if (position != currentPosition)
				{
					if (currentPosition != -1) 
					{
						out.basePrintData(data);
						Arrays.fill(data, null);
						++numDone;
						if (progress != null && numDone % progressInterval == 0) 
						{
							int percent = count == 0 ? 0 : 10+(int)((90L * numDone) / count);
							progress.display(percent, "Exporting spot data: " + 
								numDone + " of " + (count == 0 ? "unknown" : count) + " done");
						}
					}
					copyReporterFields(position, data, 0);
					currentPosition = position;
				}
				int index = numReporterFields;
				if (numAssaysPerRow > 1)
				{
					short column = result.getShort(colIndex);
					index = getAssayIndex(column);
				}
				for (int i = 1; i <= numSpotFields; ++i)
				{
					data[index] = formatters[i].format(result.getObject(i));
					++index;
				}
			}
			// Print the final line
			out.basePrintData(data);
			out.flush();
		}
		catch (SQLException ex)
		{
			throw new DatabaseException(ex);
		}
		finally
		{
			if (it != null) it.close();
		}
	}
	
	/**
		Get a configured query that counts the number of data lines that is about to
		be written in a data section. The returned query finds this value by calculating
		the number of unique positions or, if average on reporters is enabled, the number
		of unique reporters.
		
		@param bioAssayRestriction If TRUE a restriction is added to the 
			query to make it return data for a single bioassay only.
			Use <code>query.setParameter("bioAssayColumn", (int)bioAssay.getDataCubeColumnNo(), Type.INT)</code> 
			to set the value for the restriction
		@return A configured query. Use {@link DynamicSpotQuery#count(DbControl)} to
			get the count
	*/
	protected DynamicSpotQuery getCountQuery(boolean bioAssayRestriction)
	{
		// Get configuration options
		BioAssaySet source = getSource();
		boolean forAverage = getAverageOnReporter();
		
		DynamicSpotQuery query = source.getSpotData();
		query.setDistinct(true);
		if (forAverage)
		{
			query.joinReporters(JoinType.INNER);
			query.select(Selects.expression(Dynamic.column(VirtualTable.POSITION, VirtualColumn.REPORTER_ID), "rep", true));
		}
		else
		{
			query.select(Selects.expression(Dynamic.column(VirtualColumn.POSITION), "pos", true));
		}
		
		// Optional restriction
		if (bioAssayRestriction)
		{
			query.restrict(
				Restrictions.eq(
					Dynamic.column(VirtualColumn.COLUMN), 
					Expressions.parameter("bioAssayColumn")
				));
		}
		return query;
	}

}
