/**
	$Id $

	Copyright (C) 2023 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions.logging;

import java.util.Deque;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.ConcurrentLinkedDeque;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import org.apache.commons.collections4.iterators.UnmodifiableIterator;

/**
	Instances hold the log entries for a single extension.

	@author nicklas
	@since 3.19.8
*/
public class ExtensionsLog 
{
	private static Map<String, ExtensionsLog> logs = new HashMap<>();
	
	/**
		Get the service log for the given extension. If the log
		doens't exists yet it can optionally be created.
	*/
	public static ExtensionsLog getLog(String extensionId, boolean create)
	{
		ExtensionsLog log = logs.get(extensionId);
		if (log == null && create)
		{
			synchronized (logs)
			{
				log = logs.get(extensionId);
				if (log == null)
				{
					log = new ExtensionsLog(extensionId);
					logs.put(extensionId, log);
				}
			}
		}
		return log;
	}
	
	/**
		Get a logger that can add entries to the log for the given extension.
		If the log doesn't exists yet, it can optionally be created.
	*/
	public static ExtensionsLogger getLogger(String extensionId, boolean create)
	{
		ExtensionsLog log = getLog(extensionId, create);
		return log == null ? null : log.createLogger();
	}

	/**
		Create a logger that can add entries to the log for the given extension.
		If the log doesn't exists yet, it is be created.
	*/
	public static ExtensionsLogger createLogger(String extensionId)
	{
		return getLog(extensionId, true).createLogger();
	}

	private final String extensionId;
	private final Deque<LogEntry> logEntries;
	private final ReadWriteLock lock;
	
	private int maxEntries;
	private volatile int numEntries;
	private volatile int numWarnings;
	private volatile int numErrors;
	
	private boolean isDebugEnabled;
	private boolean copyToStdout;
	
	private ExtensionsLog(String extensionId)
	{
		this.extensionId = extensionId;
		this.logEntries = new ConcurrentLinkedDeque<>();
		this.maxEntries = 200;
		this.lock = new ReentrantReadWriteLock();
	}
	
	/**
		Get the extension id of the log.
	*/
	public String getExtensionId()
	{
		return extensionId;
	}
	
	/**
		Is the debug-level enabled?
	*/
	public boolean isDebugEnabled()
	{
		return isDebugEnabled;
	}
	
	/**
		Enable or disable debug-level logging.
	*/
	public void setDebugEnabled(boolean debugEnabled)
	{
		this.isDebugEnabled = debugEnabled;
	}
	
	/**
		Are log entries copied to standard output?
	*/
	public boolean isCopyingToStdout()
	{
		return copyToStdout;
	}
	
	/**
		Enable or disable copies of log entries to standard output.
		When enabled, TRACE-level log entries are also written if
		debugging is enabled.
	*/
	public void setCopyToStdout(boolean copyToStdout)
	{
		this.copyToStdout = copyToStdout;
	}
	
	/**
		Get the max number of entries to keep in the log.
	*/
	public int getMaxEntries()
	{
		return maxEntries;
	}
	
	/**
		Get the current number of entries in the log.
	*/
	public int getNumEntries()
	{
		return numEntries;
	}
	
	/**
		Get the current number of warning entries in the log.
	*/
	public int getNumWarnings()
	{
		return numWarnings;
	}
	
	/**
		Get the current number of error entries in the log.
	*/
	public int getNumErrors()
	{
		return numErrors;
	}
	
	/**
		Clear the log.
	*/
	public void clearLog()
	{
		// We need the write lock here since we are changing multiple properties
		Lock l = lock.writeLock();
		try
		{
			l.lock();
			logEntries.clear();
			numEntries = 0;
			numWarnings = 0;
			numErrors = 0;
		}
		finally
		{
			l.unlock();
		}
	}
	
	/**
		Create a logger that adds messages to this log.
	*/
	public ExtensionsLogger createLogger()
	{
		return new ExtensionsLogger(this);
	}

	/**
		Add a log entry.
		DEBUG level messages are skipped unless debug-level is enabled.
		TRACE level messages are always skipped but can be written to
		standard output if that option and debugging is enabled.
	*/
	void add(LogEntry entry)
	{
		if (copyToStdout && (entry.getLevel() != LogLevel.TRACE || isDebugEnabled))
		{
			System.out.println("["+extensionId+"] "+entry);
		}
		if (entry.getLevel() == LogLevel.TRACE || (entry.getLevel() == LogLevel.DEBUG && !isDebugEnabled)) 
		{
			return;
		}
		// A read lock should be enough since we are using a ConcurrentLinkedDeque that supports multiple threads
		Lock l = lock.readLock();
		try
		{
			l.lock();
			logEntries.addFirst(entry);
			if (entry.getLevel() == LogLevel.WARNING) numWarnings++;
			if (entry.getLevel() == LogLevel.ERROR) numErrors++;
			
			if (++numEntries > maxEntries)
			{
				// Remove one entry
				numEntries--;
				LogEntry last = logEntries.removeLast();
				if (last.getLevel() == LogLevel.WARNING) numWarnings--;
				if (last.getLevel() == LogLevel.ERROR) numErrors--;
			}
		}
		finally
		{
			l.unlock();
		}
	}
	
	/**
		Get an iterator that is returning all entries in the log.
		The returned iterator is read-only and the {@link Iterator#remove()} 
		method will throw an exception.
	*/
	public Iterator<LogEntry> getLogEntries()
	{
		return UnmodifiableIterator.unmodifiableIterator(logEntries.iterator());
	}

}
