/**
	$Id: DeleteResourcesProcessor.java 6444 2014-04-09 12:21:20Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions.manager.processor;

import java.io.File;

import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.extensions.manager.ExtensionsFile;
import net.sf.basedb.util.extensions.manager.ExtensionsFileProcessor;
import net.sf.basedb.util.extensions.manager.ExtensionsManager;
import net.sf.basedb.util.extensions.manager.ProcessResults;
import net.sf.basedb.util.extensions.manager.ExtensionsFile.WriteableExtensionsFile;
import net.sf.basedb.util.extensions.manager.filter.DeletedFilter;

/**
	A file processor implementation that removes all resources that 
	was previously extracted from a JAR file. The main target directory must 
	be specified when an instance is constructed. Each processed JAR file is 
	assumed to have been extracted into a subdirectory with the same name
	as the JAR file. This directory will be removed.
	<p>
	This processor is usually paired with a {@link DeletedFilter} to only
	delete resources related to extension files that have been deleted. 
	Using this processor without a filter will remove resources for
	all extensions.
	<p>
	Extensions files that are not a JAR file are always ignored by
	this processer.

	@author Nicklas
	@since 3.0
	@base.modified $Date: 2014-04-09 14:21:20 +0200 (Wed, 09 Apr 2014) $
*/
public class DeleteResourcesProcessor
	implements ExtensionsFileProcessor
{

	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(DeleteResourcesProcessor.class);
	
	private final File mainDir;
	private final ProcessResults results;
	
	private int numFiles;
	private int numError;
	private int numDeleted;

	
	/**
		Create a new processor.
		
		@param mainDir The main directory to where the resources are located
		@param results Optional object for storing results
	*/
	public DeleteResourcesProcessor(File mainDir, ProcessResults results)
	{
		this.mainDir = mainDir;
		this.results = results;
	}
	
	/*
		From the ExtensionsFileProcessor interface
		------------------------------------------
	*/
	@Override
	public void begin(ExtensionsManager manager, int numFiles)
	{
		this.numFiles = 0;
		this.numError = 0;
		this.numDeleted = 0;
	}

	@Override
	public void processFile(ExtensionsManager manager, WriteableExtensionsFile wFile)
	{
		ExtensionsFile xtFile = wFile.getExtensionsFile();
		if (!xtFile.isJar()) 
		{
			log.info("File is not a JAR file (skipping): " + xtFile);
			return;
		}
		
		File jar = xtFile.getFile();
		File targetDir = new File(mainDir, jar.getName());
		log.info("Removing resources from directory: " + targetDir);
		try
		{
			int num = FileUtil.deleteTempDirectory(targetDir);
			numDeleted += num;
			numFiles++;
			if (results != null)
			{
				results.addMessage(xtFile, num + " resources deleted.");
			}
			log.info("Removed  " + num + " resources from directory: " + targetDir);
		}
		catch (Throwable ex)
		{
			wFile.setError(true);
			numError++;
			if (results != null)
			{
				results.addErrorMessage(xtFile, "Could remove resources: " + ex.getClass().getName() + "; " + ex.getMessage());
			}
			log.error("Could not remove resources in directory: " + targetDir, ex);
		}
	}
	
	@Override
	public void done(ExtensionsManager manager)
	{}
	
	@Override
	public void done(ExtensionsManager manager, Throwable t)
	{}
	// ---------------------------------------------------
	
	/**
		Get the number of files that was successfully processed.
	*/
	public int getNumFiles()
	{
		return numFiles;
	}
	
	/**
		Get the number of files that was had an error.
	*/
	public int getNumError()
	{
		return numError;
	}
	
	/**
		Get the total number of resources that was deleted.
	*/
	public int getNumDeleted()
	{
		return numDeleted;
	}

	
}
