/**
	$Id: ValidationRenderer.java 8117 2023-02-22 07:58:38Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.fileset;

import java.util.ArrayList;
import java.util.List;

import net.sf.basedb.core.InvalidRelationException;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.extensions.AbstractRenderer;
import net.sf.basedb.util.extensions.Extension;

/**
	Renderer implementation for validating the members of a file set.
	
	@author Nicklas
	@since 3.0
	@base.modified $Date: 2023-02-22 08:58:38 +0100 (Wed, 22 Feb 2023) $
*/
public class ValidationRenderer
	extends AbstractRenderer<ValidationAction>
{

	private final List<ValidatingFileSetMember> members;
	private final List<Throwable> errors;
	
	private boolean isValid;
	private String message;
	
	/**
		Create a new renderer instance.
	
		@param members A list with the members of the file set
		@param errors A list for collecting validation errors
	*/
	public ValidationRenderer(List<ValidatingFileSetMember> members, List<Throwable> errors)
	{
		this.members = members;
		this.errors = errors;
	}

	/*
		From the Renderer interface
		---------------------------
	*/
	@Override
	public void render(ValidationAction action, Extension<? extends ValidationAction> ext)
	{
		ThreadSignalHandler.checkInterrupted();
		
		// Check if the action accepts any file(s)
		List<ValidatingFileSetMember> accepted = new ArrayList<ValidatingFileSetMember>();
		for (ValidatingFileSetMember member : members)
		{
			Accept accept = member.isAcceptedByValidator(action);
			if (accept == Accept.VALIDATE_IMMEDIATELY)
			{
				// Validate the file immediately
				validate(action);
				member.setResult(isValid, message);
			}
			else if (accept == Accept.VALIDATE_LATER)
			{
				// File will be validated in batch later
				accepted.add(member);
			}
		}
		
		// Are there any files waiting for validation?
		if (accepted.size() == 0) return;

		// Validate the remaining files
		validate(action);
		for (ValidatingFileSetMember member : accepted)
		{
			member.setResult(isValid, message);
		}
	}
	// -----------------------------------
	
	private void validate(ValidationAction action)
	{
		try
		{
			action.validateAndExtractMetadata();
			isValid = true;
			message = null;
		}
		catch (Throwable t)
		{
			boolean invalidRelation = t instanceof InvalidRelationException;
			boolean ignore = t instanceof PermissionDeniedException;
			isValid = invalidRelation;
			message = t.getMessage();
			if (!ignore)
			{
				errors.add(t);
				if (!invalidRelation) action.resetMetadata();
			}
		}
	}
	
}
