/*
	$Id: MeanFunction.java 6875 2015-04-20 09:08:18Z nicklas $

	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Gregory Vincic

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.jep;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.DynamicResultIterator;
import net.sf.basedb.core.DynamicRawDataQuery;
import net.sf.basedb.core.RawDataUtil;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.data.RawData;
import net.sf.basedb.core.query.Aggregations;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Selects;

import java.util.Stack;
import java.util.Map;
import java.util.HashMap;
import java.sql.SQLException;

import org.nfunk.jep.Node;
import org.nfunk.jep.ParseException;


/**
	A JEP function class that adds a <code>mean(string)</code> function to a 
	JEP expression parser. The function will calculate the mean value of the raw data 
	property with the given name. For example: 
	<code>mean('ch1BgMean')</code>
	<p>
	To be able to use this function it must be registered with the JEP
	parser and, before the expression is evaluated, a rawbioassay object must be set.
	For example we can evaluate an expression for every raw data object
	in a raw bioassay:
	
	<pre class="code">
DbControl dc = ...
RawBioAssay assay = ...
String expression = "raw('ch1FgMean') - mean('ch1BgMean')";
RawFunction raw = new RawFunction(assay.getRawDataType());
MeanFunction mean = new MeanFunction(assay);
JEP jep = JepUtil.newJep(expression, raw, mean);
DataResultIterator&lt;RawData&gt; result = 
   assay.getRawData().iterate(dc);
while (result.hasNext())
{
   raw.setRawData(result.next());
   double value = jep.getValue();
   // Do something with the value
}
result.close();
</pre>

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2015-04-20 11:08:18 +0200 (Mon, 20 Apr 2015) $
	@see Jep
*/
public class MeanFunction
	implements JepExpressionFunction
{
	/**
		Holds a cache of the values.
	*/
	private final Map<String, Float> means;
	private final DbControl dc;
	private int numParameters;
	private RawBioAssay rawBioAssay;
	private RawData rawData;

	/**
		Create a new instance of this function working with raw data of
		the specified raw bio assay.
		@param dc DbControl used in this function
	*/
	public MeanFunction(DbControl dc)
	{
		this.dc = dc;
		this.means = new HashMap<String, Float>();
	}
	
	/*
		From the JepFunction interface
		-------------------------------------------
	*/
	/**
		@return The string "mean"
	*/
	@Override
	public String getFunctionName()
	{
		return "mean";
	}
	// -------------------------------------------
	/*
		From the JepExpressionFunction interface
		-------------------------------------------
	*/
	/**
		Use the {@link Dynamic#meanRawData(String)} method to create an
		expression for the mean of the given raw data property.
	*/
	@Override
	public Expression toExpression(Node node)
	{
		int numChildren = node.jjtGetNumChildren();
		if (numChildren != 1)
		{
			throw new BaseException("Invalid number of arguments for 'mean' function: " + numChildren);
		}
		return Dynamic.meanRawData(Jep.nodeToString(node.jjtGetChild(0)));
	}
	// -------------------------------------------
	/*
		From the PostfixMathCommandI interface
		-------------------------------------------
	*/
	/**
		@return Always 1
	*/
	@Override
	public int getNumberOfParameters()
	{
		return 1;
	}
	@Override
	public void setCurNumberOfParameters(int n)
	{
		this.numParameters = n;
	}
	@Override
	public boolean checkNumberOfParameters(int n)
	{
		return n == 1;
	}
	@Override
	@SuppressWarnings({"unchecked", "rawtypes"})
	public void run(Stack stack)
		throws ParseException
	{
		if (stack == null || stack.empty()) 
		{
			throw new ParseException("Stack is empty");
		}
		Object propertyName = stack.pop();
		if (propertyName instanceof String)
		{
			stack.push(mean((String)propertyName));
		}
		else
		{
			throw new ParseException("Invalid parameter type: " + propertyName + "; expected string");
		}
	}
	// -------------------------------------------

	/**
		Set a new raw data object that will be used the next time the
		JEP expression is evaluated.
		@param rawData The raw data object to use
	*/
	public void setRawData(RawData rawData)
	{
		this.rawData = rawData;
	}
	/**
		Get the current raw data object used when evaluating the JEP expression.
	*/
	public RawData getRawData()
	{
		return rawData;
	}

	
	/**
		Get the value of the specified property of the current raw data object.
		This method uses Hibernate metadata to find the value.
		@param propertyName The name of the raw data property. Null is not allowed.
	 	@return The mean value as a float. 
	  	@throws ParseException If the property name is null or 
	   		if there is another error.
	*/
	public float mean(String propertyName)
		throws ParseException
	{
		if (propertyName == null)
		{
			throw new ParseException("No property name has been specified for function mean()");
		}
		if (rawData == null)
		{
			throw new ParseException("No raw data object has been specified for function mean('"+propertyName+"')");
		}
		Float theMean = null;
		try
		{
			RawBioAssay rba = RawDataUtil.getRawBioAssay(dc, rawData);
			if (rawBioAssay == null || rba.getId() != rawBioAssay.getId())
			{
				// Raw data from is from a different RawBioAssay. Clean cache.
				rawBioAssay = rba;
				means.clear();
			}
			theMean = means.get(propertyName);
			if (theMean == null)
			{
				// Value not found in cache, query the database
				DynamicRawDataQuery rawQuery = rawBioAssay.getDynamicRawData();
				rawQuery.select(Selects.expression(Aggregations.mean(Dynamic.rawData(propertyName)), "mean"));
				DynamicResultIterator result = rawQuery.iterate(rawBioAssay.getDbControl());
				theMean = result.next().getFloat(1);
				means.put(propertyName, theMean);
			}
		}
		catch (SQLException ex)
		{
			throw new ParseException(ex.getMessage());
		}
		catch (BaseException ex)
		{
			throw new ParseException(ex.getMessage());
		}
		return theMean;
	}
}
