/**
	$Id$

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.listable;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.Listable;
import net.sf.basedb.core.SyncFilter.SourceItemTransform;

/**
	Utility class for working with {@link Listable} items.
	@author nicklas
	@since 3.5
*/
public class ListableUtil 
{


	/**
		Get all item types that can have the specified relation to the 
		with the given item type. Depending on the transform parameter
		this return either the parent or child items or the item itself.
	*/
	public static List<Item> getListableItems(Item item, SourceItemTransform transform)
	{
		List<Item> items = null;
		if (transform != SourceItemTransform.NONE)
		{
			SourceItemTransformerFactory factory = getTransformerFactory(item);
			if (factory != null) items = factory.getSupportedSourceItems(transform);
		}
		else if (Listable.class.isAssignableFrom(item.getItemClass()))
		{
			items = Arrays.asList(item);
		}
		if (items == null) items = Collections.emptyList();
		return items;
	}
	
	/**
		Create a source item transformer factory that can transform items to the given
		target item type. If no transformer factory exists, null is returned.
	*/
	public static SourceItemTransformerFactory getTransformerFactory(Item targetItemType)
	{
		return getTransformerFactory(targetItemType, false);
	}
	
	/**
		Create a source item transformer factory that can transform items to the given
		target item type. If no transformer factory exists, null is returned.
		The 'includeChildrenThatPushToParent' parameter can be used for transforms
		that go from child to parent items, when the target is a SAMPLE or EXTRACT.
		When this flag is set, transformer will also include child items that have
		a subtype that have the {@link ItemSubtype#getPushAnnotations()} flag enabled
		in the final result.
		
		@since 3.19.5
	*/
	public static SourceItemTransformerFactory getTransformerFactory(Item targetItemType, boolean includeChildrenThatPushToParent)
	{
		SourceItemTransformerFactory factory = null;
		if (targetItemType == Item.BIOSOURCE)
		{
			factory = new ToBioSourceSourceItemTransformerFactory();
		}
		else if (targetItemType == Item.SAMPLE)
		{
			factory = new ToSampleSourceItemTransformerFactory(includeChildrenThatPushToParent);
		}
		else if (targetItemType == Item.EXTRACT)
		{
			factory = new ToExtractSourceItemTransformerFactory(includeChildrenThatPushToParent);
		}
		else if (targetItemType == Item.PHYSICALBIOASSAY)
		{
			factory = new ToPhysicalBioAssaySourceItemTransformerFactory();
		}
		else if (targetItemType == Item.DERIVEDBIOASSAY)
		{
			factory = new ToDerivedBioAssaySourceItemTransformerFactory();
		}
		else if (targetItemType == Item.RAWBIOASSAY)
		{
			factory = new ToRawBioAssaySourceItemTransformerFactory();
		}
		return factory;
	}
}
