/**
	$Id$

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.listable;

import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DerivedBioAssay;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

/**
	Source item transformer implementation that transform physical bioassay items 
	to derived bioassays. This implementation transform to all derived bioassays 
	that are descendants of the physical bioassays and may optionally limit the
	derived bioassays that are related to parent collected extracts.
	
	@author Nicklas
	@since 3.5
*/
public class PhysicalBioAssayToDerivedBioAssayTransformer 
	extends AbstractSourceItemTransformer
{

	private CollectExtracts collectedExtracts;
	
	/**
		Create a new physical bioassay to derived bioassay transformer.
	*/
	public PhysicalBioAssayToDerivedBioAssayTransformer() 
	{
		super(Item.PHYSICALBIOASSAY, Item.DERIVEDBIOASSAY);
	}
	
	/**
		Create a new physical bioassay to derived bioassay transformer using
		the collected extracts to limit the returned items.
	*/
	public PhysicalBioAssayToDerivedBioAssayTransformer(CollectExtracts collectedExtracts) 
	{
		this();
		this.collectedExtracts = collectedExtracts;
	}
	
	@Override
	public Set<Integer> transform(TransformContext context, Set<Integer> source) 
	{
		DbControl dc = context.getDbControl();
		
		ItemQuery<DerivedBioAssay> query = DerivedBioAssay.getQuery();
		query.setIncludes(context.getInclude());
		query.join(Hql.innerJoin("physicalBioAssays", "pba"));
		query.restrict(
			Restrictions.in(
				Hql.alias("pba"), 
				Expressions.parameter("bioAssays")
			)
		);
		context.collect(getSourceItemType(), source);
		Set<Integer> all = context.avoid(getTargetItemType(), safeIdList(dc, query, "bioAssays", source));
		
		if (collectedExtracts != null)
		{
			// Keep only derived bioassays that have a 'null' extract or
			// an extract that is included among the collected or their parents
			Set<Integer> collectedAndParents; 
			if (collectedExtracts.getCollectedParents().size() > 0)
			{
				// merge
				collectedAndParents = new HashSet<Integer>(collectedExtracts.getCollected());
				collectedAndParents.addAll(collectedExtracts.getCollectedParents());
			}
			else
			{
				// only use the collected
				collectedAndParents = collectedExtracts.getCollected();
			}
			
			query = DerivedBioAssay.getQuery();
			query.setIncludes(context.getInclude());
			query.join(Hql.leftJoin("extract", "xtr"));
			query.restrict(Restrictions.or(
				Restrictions.eq(Hql.alias("xtr"), null),
				Restrictions.in(Hql.alias("xtr"), Expressions.parameter("extracts"))
			));
			all.retainAll(safeIdList(dc, query, "extracts", collectedAndParents));
		}
		
		return all;
	}

}
