/**
	$Id$

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.listable;

import java.util.Collections;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.Item;
import net.sf.basedb.core.SyncFilter.SourceItemTransform;
import net.sf.basedb.util.listable.TransformCache.CacheKey;

/**
	This is a wrapper transformer factory that can be applied to any other
	factory in order to provide support for caching.
	
	The cache holds the result from {@link SourceItemTransformer#transform(TransformContext, Set)}
	so that if the same set of source ids are provided it will pick the cache result if
	available and only forward to the wrapped factory in case no old result exists or if it
	is too old.

	@author Nicklas
	@since 3.16
*/
public class SourceItemTransformerWithCache 
	implements SourceItemTransformerFactory 
{

	private final SourceItemTransformerFactory factory;
	private final String cacheRegion;
	
	/**
		Create a new source item transformer chain. The chain
		must contain at least one element and all steps must have
		matching source and target item types.
		@throws IllegalArgumentException If the chain is null or empty or have non-matching
			transformation steps
	*/
	public SourceItemTransformerWithCache(SourceItemTransformerFactory factory)
	{
		this(factory, null);
	}

	public SourceItemTransformerWithCache(SourceItemTransformerFactory factory, String cacheRegion)
	{
		this.factory = factory;
		this.cacheRegion = cacheRegion;
	}
	
	
	@Override
	public Item getTargetItem() 
	{
		return factory.getTargetItem();
	}

	@Override
	public List<Item> getSupportedSourceItems(SourceItemTransform transform) 
	{
		return factory.getSupportedSourceItems(transform);
	}

	@Override
	public SourceItemTransformer create(Item sourceItemType, SourceItemTransform transform) 
	{
		return new TransformerWithCache(factory.create(sourceItemType, transform), transform, cacheRegion);
	}
	

	static class TransformerWithCache
		implements SourceItemTransformer
	{

		private final SourceItemTransformer transformer;
		private final SourceItemTransform transform;
		private final String cacheRegion;
		
		TransformerWithCache(SourceItemTransformer transformer, SourceItemTransform transform, String cacheRegion)
		{
			this.transformer = transformer;
			this.transform = transform;
			this.cacheRegion = cacheRegion;
		}
		
		@Override
		public Item getSourceItemType() 
		{
			return transformer.getSourceItemType();
		}

		@Override
		public Item getTargetItemType() 
		{
			return transformer.getTargetItemType();
		}

		/**
			If a cache exists check if it has data for this transform already.
			If not, forward to the parent transformer and then cache the result
			for future use.
		*/
		@Override
		public Set<Integer> transform(TransformContext context, Set<Integer> source) 
		{
			Set<Integer> target = null;
			TransformCache cache = context.getCache();
			CacheKey key = null;
			if (cache != null)
			{
				key = TransformCache.getKey(cacheRegion, getSourceItemType(), getTargetItemType(), transform, source);
				target = cache.get(key);
			}
			if (target == null)
			{
				target = transformer.transform(context, source);
				if (cache != null) cache.store(key, target);
			}
			return Collections.unmodifiableSet(target);
		}
	}
}
