/**
	$Id$

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.listable;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.Item;

/**
	Transformation context that can be used to keep track of vital
	information when doing multi-step transformations.

	@author Nicklas
	@since 3.5
*/
public class TransformContext 
{

	private final DbControl dc;
	private Collection<Include> include;
	
	/**
		Creates a new transformation context that will use the given DbControl
		for database access.
	*/
	public TransformContext(DbControl dc)
	{
		this.dc = dc;
		this.include = Include.ALL;
	}
	
	/**
		Get the DbControl to use in this transformation context.
	*/
	public DbControl getDbControl()
	{
		return dc;
	}
	
	/**
		Include options when executing queries to find
		parent/child items. Default is Include.ALL.
	*/
	public Collection<Include> getInclude()
	{
		return include;
	}
	
	public void setInclude(Collection<Include> include)
	{
		this.include = include;
	}
	
	/**
		Get a cache that can be used for storing transformation
		results for some time. The cache is stored in the current
		session and is lost if the user logs out.
		@since 3.16
	*/
	public TransformCache getCache()
	{
		TransformCache cache = dc.getSessionControl().getSessionSetting(TransformCache.class.getName());
		if (cache == null)
		{
			cache = new TransformCache(20);
			dc.getSessionControl().setSessionSetting(TransformCache.class.getName(), cache);
		}
		return cache;
	}

	private boolean collecting;
	private boolean avoiding;
	private Map<Item, Set<Integer>> collectedByItemType;

	
	/**
		Enable/disable collecting of all source items that the transformers in
		this context are using.
		@since 3.19.5
	*/
	public void setCollecting(boolean collecting)
	{
		this.collecting = collecting;
	}
	
	/**
		Shortcut for enabling 'collecting' and disabling 'avoiding'.
		@since 3.19.5
	*/
	public void setCollecting()
	{
		this.collecting = true;
		this.avoiding = false;
	}
	
	public boolean isCollecting()
	{
		return collecting;
	}
	
	/**
		Enable/disable avoiding of returning any source items that has 
		been previously collected.
		@since 3.19.5
	*/
	public void setAvoiding(boolean avoiding)
	{
		this.avoiding = avoiding;
	}
	/**
		Shortcut for enabling 'avoiding' and disabling 'collecting'.
		@since 3.19.5
	*/
	public void setAvoiding()
	{
		this.collecting = false;
		this.avoiding = true;
	}

	public boolean isAvoiding()
	{
		return avoiding;
	}
	
	/**
		Clear all collected items and disable both 'collecting' and
		'avoiding'.
		@since 3.19.5
	*/
	public void resetCollected()
	{
		this.collecting = false;
		this.avoiding = false;
		if (collectedByItemType != null) collectedByItemType.clear();
	}
	
	/**
		Store collected items for later use if collecting has been enabled.
		@return The source items
	*/
	public Set<Integer> collect(Item itemType, Set<Integer> source)
	{
		if (collecting && source.size() > 0)
		{
			if (collectedByItemType == null) collectedByItemType = new HashMap<>();
			Set<Integer> collected = collectedByItemType.get(itemType);
			if (collected == null) 
			{
				collected = new HashSet<>();
				collectedByItemType.put(itemType, collected);
			}
			collected.addAll(source);
		}
		return source;
	}
	
	/**
		Avoid collected items by removing them from the source set.
		Note that this method modified the source set directly.
		@return The source set
	*/
	public Set<Integer> avoid(Item itemType, Set<Integer> source)
	{
		if (avoiding && source.size() > 0)
		{
			if (collectedByItemType != null && collectedByItemType.containsKey(itemType))
			{
				Set<Integer> collected = collectedByItemType.get(itemType);
				source.removeAll(collected);
			}
		}
		return source;
	}

}
