/**
	$Id: HardwareLoader.java 6870 2015-04-16 11:47:41Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;

import net.sf.basedb.core.ArrayBatch;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.BioPlate;
import net.sf.basedb.core.BioPlateEvent;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DerivedBioAssay;
import net.sf.basedb.core.Hardware;
import net.sf.basedb.core.MeasuredBioMaterial;
import net.sf.basedb.core.PhysicalBioAssay;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.node.ChildNodeDirection;
import net.sf.basedb.util.overview.node.NameableNameGenerator;
import net.sf.basedb.util.overview.node.NodeFactory;

/**
	Node loader implementation for hardware. Hardware is loaded as
	end-point property nodes with no children. It can't be a root 
	node or a forward/reverse-loading node.
	<p>
	Validations:
	<ul>
	<li>Missing hardware: {@link Validator#MISSING_HARDWARE}
	<li>Denied hardware: {@link Validator#DENIED_HARDWARE}
	<li>Incorrect hardware type: {@link Validator#INCORRECT_HARDWARETYPE} 
	<li>Non-default hardware: {@link Validator#NONDEFAULT_HARDWARE} 
	</ul>
	
	@author Nicklas
	@version 2.10
	@base.modified $Date: 2015-04-16 13:47:41 +0200 (Thu, 16 Apr 2015) $
*/
public class HardwareLoader
	extends BasicItemNodeLoader<Hardware>
{
	
	public HardwareLoader()
	{
		super(Item.HARDWARE, DENY_ROOT_NODE, 
				new NameableNameGenerator<Hardware>("hardware", "Hardware"));
	}
	
	/*
		From the NodeLoader interface
		------------------------------
	*/
	@Override
	public Node createPropertyNode(DbControl dc, OverviewContext context, Node parentNode)
	{
		Hardware hardware = null;
		boolean denied = false;
		try
		{
			BasicItem parentItem = parentNode.getItem(dc);
			if (parentItem instanceof DerivedBioAssay)
			{
				hardware = ((DerivedBioAssay)parentItem).getHardware();
			}
			else if (parentItem instanceof PhysicalBioAssay)
			{
				hardware = ((PhysicalBioAssay)parentItem).getCreationEvent().getHardware();
			}
			else if (parentItem instanceof MeasuredBioMaterial)
			{
				hardware = ((MeasuredBioMaterial)parentItem).getCreationEvent().getHardware();
			}
			else if (parentItem instanceof ArrayBatch)
			{
				hardware = ((ArrayBatch)parentItem).getPrintRobot();
			}
			else if (parentItem instanceof BioPlate)
			{
				hardware = ((BioPlate)parentItem).getFreezer();
			}
			else if (parentItem instanceof BioPlateEvent)
			{
				hardware = ((BioPlateEvent)parentItem).getHardware();
			}
			// If there are other items with a software
			// we should add more cases here
		}
		catch (PermissionDeniedException ex)
		{
			denied = true;
		}
		
		NodeFactory<Hardware> nf = getNodeFactory(dc, context);
		Node hardwareNode = createItemNode(nf, hardware, hardware, denied, 
				parentNode, ChildNodeDirection.PROPERTY);
		return hardwareNode;
	}
	// -----------------------------------
	/*
		From the AbstractNodeLoader class
		----------------------------------
	*/
	/**
		Loads annotations and hardware parameters for the given hardware node.
		@see RawBioAssayLoader#createForwardNode(DbControl, OverviewContext, Node)
	*/
	@Override
	protected void loadPropertyChildNodes(DbControl dc, OverviewContext context, Node hardwareNode)
	{
		getNodeLoader(context, Item.ANNOTATION).createPropertyNode(dc, context, hardwareNode);
	}
	// -------------------------------------	
}
