/**
	$Id: AbstractFileUnpacker.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) 2006, 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.zip;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.AbsoluteProgressReporter;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Directory;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.PathParameterType;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.SimpleAbsoluteProgressReporter;
import net.sf.basedb.core.plugin.AbstractPlugin;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Plugin;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalTarget;
import net.sf.basedb.core.signal.ThreadSignalHandler;

/**
 	This is a base class for all plugins that wants to implement the {@link FileUnpacker}
 	interface. This class takes care of all the hard work that is required by a plugin. 
 	Ie. the creating and saving the parameters that are needed, context checking, etc.
 	The subclass only has to implement the remaingin methods in the {@link FileUnpacker}
 	interface which basically boils down to implementing the actual unpacking of the
 	compressed file.
 	<p>
 	The implementation in this class asks for three parameters:
 	<ul>
 	<li>The file to unpack
 	<li>The directory where the files should be placed
 	<li>If existing files should be overwritten or not
 	</ul>
 	A subclass may override any of the methods in this class if it has different
 	needs.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2012-09-14 09:28:35 +0200 (Fri, 14 Sep 2012) $
*/
public abstract class AbstractFileUnpacker
	extends AbstractPlugin
	implements FileUnpacker, InteractivePlugin, SignalTarget
{

	/**
		The context is always [ FILE, ITEM].
	*/
	private static final Set<GuiContext> guiContexts = 
		Collections.singleton(new GuiContext(Item.FILE, GuiContext.Type.ITEM));

	private static final Set<Permissions> permissions = new HashSet<Permissions>();

	/**
		The parameter type for file to unpack.
	*/
	private static final ItemParameterType<File> zipFileType = 
		new ItemParameterType<File>(File.class, null, true, 1, null);
	
	/**
		The parameter type for unpack directory.
	*/
	private static final PathParameterType unpackType =
		new PathParameterType(Path.Type.DIRECTORY, null, true);
	
	private RequestInformation configureJob;
	
	private ThreadSignalHandler signalHandler;
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	@Override
	public MainType getMainType()
	{
		return Plugin.MainType.OTHER;
	}
	/**
		Returns FALSE.
	*/
	@Override
	public boolean supportsConfigurations()
	{
		return false;
	}
	/**
		Returns FALSE.
	*/
	@Override
	public boolean requiresConfiguration()
	{
		return false;
	}
	/**
	 	Request create and write access to File:s and Directory:s.
	*/
	@Override
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			permissions.add(new Permissions(Item.FILE, null, EnumSet.of(Permission.CREATE, Permission.WRITE)));
			permissions.add(new Permissions(Item.DIRECTORY, null, EnumSet.of(Permission.CREATE, Permission.WRITE)));
			permissions.add(new Permissions(Item.ITEMSUBTYPE, EnumSet.of(Permission.USE), null));
		}
		return permissions;
	}
	
	@Override
	public void run(Request request, Response response, ProgressReporter progress)
	{
		DbControl dc = sc.newDbControl();
		InputStream in = null;
		try
		{
			// Get all parameters
			File zipFile = getZipFile(dc);
			Directory dir = getUnpackDirectory(dc);
			boolean overwrite = getOverwrite();
			
			in = zipFile.getDownloadStream(0);
			AbsoluteProgressReporter absoluteProgress = progress == null ? 
				null : new SimpleAbsoluteProgressReporter(progress, zipFile.getSize());
			
			// Unpack the file
			int numFiles = unpack(dc, dir, in, zipFile, overwrite, absoluteProgress);
			dc.commit();
			
			response.setDone(numFiles + " file(s) unpacked successfully");
		}
		catch (Throwable t)
		{
			response.setError(t.getMessage(), Arrays.asList(t));
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	// -------------------------------------------

	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	@Override
	public String isInContext(GuiContext context, Object item)
	{
		String message = null;	
		if (item == null)
		{
			message = "The object is null";
		}
		else if (!(item instanceof File))
		{
			message = "The object is not a File: " + item;
		}
		else
		{
			File f = (File)item;
			String name = f.getName();
			String mimeType = f.getMimeType();
			if (!getMimeTypes().contains(mimeType))
			{
				for (String extension : getExtensions())
				{
					if (name.endsWith(extension)) return null;
				}
				message = "Unsupported file: " + f.getPath().toString();
			}
		}
		return message;
	}

	@Override
	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJobParameters(context);
		}
		return requestInformation;
	}
	@Override
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();		
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				RequestInformation ri = getConfigureJobParameters(context);
				List<Throwable> errors = validateRequestParameters (ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size() + " invalid parameters were found in the request",errors);
					return;
				}
				
				storeValue(job, request, ri.getParameter(ZIP_FILE));
				storeValue(job, request, ri.getParameter(UNPACK_DIRECTORY));
				storeValue(job, request, ri.getParameter(OVERWRITE));
				
				File file = (File)job.getValue(ZIP_FILE);
				response.setSuggestedJobName("Unpack file '" + file.getName() + 
						"' to " + job.getValue(UNPACK_DIRECTORY));
				
				response.setDone("The job configuration is complete");
			}
		}
		catch(Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}
	// -------------------------------------------
	/*
		From the SignalTarget interface
		-------------------------------------------
	*/
	@Override
	public SignalHandler getSignalHandler()
	{
		signalHandler = new ThreadSignalHandler();
		return signalHandler;
	}
	// -------------------------------------------

	/**
		The name of the parameter that asks for the file that is going to
		be unpacked.
		@see #getZipFileParameter(String, String)
	*/
	protected static final String ZIP_FILE = "zipFile";
	
	/**
		Get a plugin parameter that asks for the file that is going to
		be unpacked.

		@param label The parameter label, or null to use the default (File)
		@param description The parameter description, or null to use the default 
			(The file to unpack.)
	*/
	protected PluginParameter<File> getZipFileParameter(String label, String description)
	{
		if (label == null) label = "File";
		if (description == null) description = "The file to unpack.";
		return new PluginParameter<File>(ZIP_FILE, label, description, zipFileType);
	}

	/**
		Get the file that is going to be unpacked by the plugin.
		The file is only available in the execution phase if the plugin has saved 
		it as a job parameter during the configuration of a job (this is the default)
		
		@return The packed file or null
		@see #getZipFileParameter(String, String)
		@see #getCurrentFile(DbControl)
	*/
	protected File getZipFile(DbControl dc)
	{
		File zipFile = (File)job.getValue(ZIP_FILE);
		if (zipFile != null) zipFile = File.getById(dc, zipFile.getId());
		return zipFile;
	}
	
	/**
		The name of the parameter that asks for the directory where the
		unpacked files are stored.
		@see #getZipFileParameter(String, String)
	*/
	protected static final String UNPACK_DIRECTORY = "unpackDirectory";

	/**
		Get a plugin parameter that asks for the directory where the
		unpacked files are stored.
	
		@param label The parameter label, or null to use the default (Save in)
		@param description The parameter description, or null to use the default 
			(The directory where the unpacked files should be saved.)
	*/
	protected PluginParameter<String> getDirectoryParameter(String label, String description)
	{
		if (label == null) label = "Save in";
		if (description == null) description = "The directory where the unpacked files should be saved.";
		DbControl dc = sc.newDbControl();
		String defaultDir = null;
		try
		{
			Directory dir = getCurrentDirectory(dc);
			if (dir != null) defaultDir = dir.getPath().toString();
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return new PluginParameter<String>(UNPACK_DIRECTORY, label, description , defaultDir, unpackType); 
	}
	
	/**
		Get the current directory or null there is no current or it can't be loaded.
		This method will first try to load the current file and use the file's
		directory as the current. If there is no current file, the current directory
		is loaded from the session context.
		The current directory is only available in the configuration phase.
		
		@param dc A DbControl to use for database access
		@return A directory or null
		@since 2.4
		@see #getUnpackDirectory(DbControl)
	*/
	protected Directory getCurrentDirectory(DbControl dc)
	{
		File currentFile = getCurrentFile(dc);
		if (currentFile != null) return currentFile.getDirectory();
		int directoryId = sc.getCurrentContext(Item.DIRECTORY).getId();
		if (directoryId != 0)
		{
			try
			{
				return Directory.getById(dc, directoryId);
			}
			catch (Throwable t)
			{}
		}
		return null;
	}
	
	/**
		Get the current file or null there is no current or it can't be loaded.
		The current file is only available in the configuration phase.
		@param dc A DbControl to use for database access
		@return A file or null
		@since 2.4
		@see #getZipFile(DbControl)
	*/
	protected File getCurrentFile(DbControl dc)
	{
		int fileId = sc.getCurrentContext(Item.FILE).getId();
		if (fileId != 0)
		{
			try
			{
				return File.getById(dc, fileId);
			}
			catch (Throwable t)
			{}
		}
		return null;
	}	
	
	/**
		Get the directory where the unpacked files are placed by the plugin.
		The directory is only available in the execution phase if the plugin has saved 
		it as a job parameter during the configuration of a job.
		
		@return The directory or null
		@see #getDirectoryParameter(String, String)
		@see #getCurrentDirectory(DbControl)
	*/
	protected Directory getUnpackDirectory(DbControl dc)
	{
		String path = (String)job.getValue(UNPACK_DIRECTORY);
		Directory unpackDirectory = null;
		if (path != null) 
		{
			unpackDirectory = Directory.getNew(dc, new Path(path, Path.Type.DIRECTORY));
			if (!unpackDirectory.isInDatabase()) dc.saveItem(unpackDirectory);
		}
		return unpackDirectory;
	}

	/**
		The name of the parameter that asks if existing files should be 
		overwritten or not.
		@see #getOverwriteParameter(String, String)
	*/
	protected static final String OVERWRITE = "overwrite";

	/**
		Get a plugin parameter that asks if existing files should be overwritten
		or not.
	
		@param label The parameter label, or null to use the default (Overwrite)
		@param description The parameter description, or null to use the default 
			(Specify if existing files should be overwritten or not.)
	*/
	protected PluginParameter<Boolean> getOverwriteParameter(String label, String description)
	{
		if (label == null) label = "Overwrite";
		if (description == null) description = "Specify if existing files should be overwritten or not.";
		return new PluginParameter<Boolean>(OVERWRITE, label, description, 
				new BooleanParameterType(false, false));
	}
	
	/**
		If existing file should be overwritten or not.
		
		@return TRUE if files should be overwritten, FALSE otherwise
		@see #getOverwriteParameter(String, String)
	*/
	protected boolean getOverwrite()
	{
		Boolean overwrite = (Boolean)job.getValue(OVERWRITE);
		if (overwrite == null) overwrite = Boolean.FALSE;
		return overwrite;
	}

	private RequestInformation getConfigureJobParameters(GuiContext context)
	{
		if (configureJob == null)
		{
			// Parameters for CONFIGURE_JOB
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();

			// Source and child bioassay set parameters
			parameters.add(getZipFileParameter(null, null));
			parameters.add(getDirectoryParameter(null, null));
			parameters.add(getOverwriteParameter(null, null));
			
			configureJob = new RequestInformation
			(
				Request.COMMAND_CONFIGURE_JOB,
				"Unpack file",
				"Select a file to unpack and where to put the unpacked files",
				parameters
			);
		}
		return configureJob;
	}
	
}
