/*
	$Id: ReporterFlatFileImporter.java 7668 2019-03-21 11:50:52Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins;

import net.sf.basedb.core.File;
import net.sf.basedb.core.HibernateUtil;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.ItemInUseException;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.ReporterList;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ExtendedProperties;
import net.sf.basedb.core.ExtendedProperty;
import net.sf.basedb.core.ReporterType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.Reporter;
import net.sf.basedb.core.ReporterBatcher;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;

import net.sf.basedb.core.data.ReporterData;

import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.plugin.GuiContext;

import net.sf.basedb.plugins.util.Parameters;
import net.sf.basedb.util.error.SimpleErrorHandler;
import net.sf.basedb.util.parser.ConfigureByExample;
import net.sf.basedb.util.parser.ConstantMapper;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;

import java.util.Collection;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.TreeSet;

/**
	A plugin that imports reporter from a flat file.
	
	@base.modified $Date: 2019-03-21 12:50:52 +0100 (Thu, 21 Mar 2019) $
	@author Nicklas
	@version 2.0
*/
public class ReporterFlatFileImporter 
	extends AbstractFlatFileImporter
	implements InteractivePlugin, ConfigureByExample
{

	private static final Set<GuiContext> guiContexts = 
		new HashSet<GuiContext>(Arrays.asList(
			new GuiContext(Item.REPORTER, GuiContext.Type.LIST),
			new GuiContext(Item.REPORTERLIST, GuiContext.Type.ITEM),
			new GuiContext(Item.REPORTERSCORE, GuiContext.Type.LIST)
		));
	
	private static final Set<Permissions> permissions = new HashSet<Permissions>();

	private static final StringParameterType requiredColumnMapping = new StringParameterType(255, null, true);
	private static final StringParameterType optionalColumnMapping = new StringParameterType(255, null, false);
	
	private static final PluginParameter<String> nameColumnMapping = new PluginParameter<String>(
		"nameColumnMapping",
		"Name",
		"Mapping that picks the reporter's name from the data columns. " +
		"For example: \\Name\\",
		requiredColumnMapping
		);
	
	private static final PluginParameter<String> reporterIdColumnMapping = new PluginParameter<String>(
		"reporterIdColumnMapping",
		"External ID",
		"Mapping that picks the reporter's external ID from the data columns. " +
		"For example: \\ID\\",
		requiredColumnMapping
		);
	
	private static final PluginParameter<String> descriptionColumnMapping = new PluginParameter<String>(
		"descriptionColumnMapping",
		"Description",
		"Mapping that picks the reporter's description from the data columns. " +
		"For example: \\Description\\",
		optionalColumnMapping
		);
	
	private static final PluginParameter<String> symbolColumnMapping = new PluginParameter<String>(
		"symbolColumnMapping",
		"Gene symbol",
		"Mapping that picks the reporter's gene symbol from the data columns. " +
		"For example: \\Gene symbol\\",
		optionalColumnMapping
		);

	private static final PluginParameter<String> scoreColumnMapping = new PluginParameter<String>(
		"scoreColumnMapping",
		"Score",
		"Mapping that picks the reporter's score in some context. This mapping is only " +
			"used when importing to a reporter list.",
		optionalColumnMapping
		);

	private static final PluginParameter<String> reporterTypeColumnMapping = new PluginParameter<String>(
		"reporterTypeColumnMapping",
		"Reporter type",
		"Mapping that pick the reporter's type from the data columns. This will overide the reporter type parameter. For example: \\Reporter type\\",
		optionalColumnMapping
		);

	protected static final PluginParameter<String> reporterIsUsedErrorParameter = new PluginParameter<String>(
			"reporterIsUsedError",
			"Reporter is used",
			"How to handle errors that are caused by a reporter beeing used by other items" +
			"and can't be deleted. This option is used in DELETE mode only. " +
			"If not specified the default error handling is used.\n\n" +
			"skip = Skip the current data line and continue\n"+
			"fail = Stop with an error message",
			new StringParameterType(255, "skip", false, 1, 0, 0, 
					Arrays.asList( new String[] { "skip", "fail"} ))
			);	
	
	private PluginParameter<ReporterType> reporterTypeParameter;
	private PluginParameter<ReporterList> reporterListParameter;
	
	private List<PluginParameter<String>> allColumnMappings;
	
	private RequestInformation configurePlugin;
	private RequestInformation configureJob;
	
	private DbControl dc;
	private ReporterBatcher batcher;
	
	/**
		Create a new importer.
	*/
	public ReporterFlatFileImporter()
	{}
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	@Override
	public boolean supportsConfigurations()
	{
		return true;
	}
	@Override
	public boolean requiresConfiguration()
	{
		return false;
	}
	/**
	 	Request create, write and delete access to Reporter:s, write access to ReporterList:s
	 	and read access to File:s and ReporterType:s.
	*/
	@Override
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			permissions.add(new Permissions(Item.REPORTER, null, EnumSet.of(Permission.CREATE, Permission.DELETE)));
			permissions.add(new Permissions(Item.REPORTERLIST, null, EnumSet.of(Permission.WRITE)));
			permissions.add(new Permissions(Item.REPORTERTYPE, EnumSet.of(Permission.CREATE, Permission.USE), null));
			permissions.add(new Permissions(Item.FILE, null, EnumSet.of(Permission.READ)));
		}
		return permissions;
	}
	// -------------------------------------------
	
	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	/**
		Return a set containing the context:s [REPORTER, LIST],
		[REPORTERLIST, ITEM], [REPORTERSCORE, LIST]. The first context
		is for importing reporters only, and the last two for importing 
		reporters to a reporter list.
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	/**
		Null if the context is {@link Item#REPORTERLIST} and the
		item is a {@link ReporterList}.
		@throws PermissionDeniedException If the user doesn't have write permission
			to the reporter list
	*/
	@Override
	public String isInContext(GuiContext context, Object item)
	{
		String message =  null;
		if (context.getItem() == Item.REPORTERLIST)
		{
			if (item == null)
			{
				message = "The object is null";
			}
			else if (!(item instanceof ReporterList))
			{
				message = "The object is not a ReporterList: " + item;
			}
			else
			{
				ReporterList rl = (ReporterList)item;
				rl.checkPermission(Permission.WRITE);
			}
		}
		return message;
	}
	/**
		The {@link Request#COMMAND_CONFIGURE_PLUGIN} command will ask for
		parser regular expressions, column mappings and the reporter type.
		The {@link Request#COMMAND_CONFIGURE_JOB} command will ask for
		a file and if existing reporters should be updated or ignored.
	*/
	@Override
	public RequestInformation getRequestInformation(GuiContext context, String command)
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
		{
			requestInformation = getConfigurePluginParameters(context);
		}
		else if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJobParameters(context);
			ItemContext fileContext = sc.getCurrentContext(Item.FILE);
			fileContext.removeAllTemporaryFilters();
		}
		return requestInformation;
	}
	/**
		Store configuration settings for {@link Request#COMMAND_CONFIGURE_PLUGIN} and
		{@link Request#COMMAND_CONFIGURE_JOB}.
	*/
	@Override
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
			{
				RequestInformation ri = getConfigurePluginParameters(context);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
					return;
				}

				// Parser settings
				storeValue(configuration, request, headerRegexpParameter);
				storeValue(configuration, request, dataHeaderRegexpParameter);
				storeValue(configuration, request, dataSplitterRegexpParameter);
				storeValue(configuration, request, trimQuotesParameter);
				storeValue(configuration, request, ignoreRegexpParameter);
				storeValue(configuration, request, dataFooterRegexpParameter);
				storeValue(configuration, request, minDataColumnsParameter);
				storeValue(configuration, request, maxDataColumnsParameter);
				storeValue(configuration, request, ri.getParameter(Parameters.CHARSET_PARAMETER));
				storeValue(configuration, request, ri.getParameter(Parameters.DECIMAL_SEPARATOR_PARAMETER));
				
				// Column mappings
				storeValue(configuration, request, complexMappings);
				boolean allowComplex = "allow".equals(request.getParameterValue(complexMappings.getName()));
				for (PluginParameter<String> pp : getAllColumnMappings(false))
				{
					String mapExpression = (String)request.getParameterValue(pp.getName());
					checkColumnMapping(mapExpression, allowComplex, pp.getLabel());
					storeValue(configuration, pp, mapExpression);
				}
				
				// Reporter type
				storeValue(configuration, request, reporterTypeParameter);
				response.setDone("Plugin configuration complete");
			}
			else if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				RequestInformation ri = getConfigureJobParameters(context);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
					return;
				}
				
				boolean reporterListContext = context != null && 
					(context.getItem() == Item.REPORTERLIST || context.getItem() == Item.REPORTERSCORE);
				if (reporterListContext)
				{
					storeValue(job, request, reporterListParameter);
				}
				storeValue(job, request, fileParameter);
				storeValue(job, request, ri.getParameter("mode"));
				
				// Parser settings
				storeValue(job, request, headerRegexpParameter);
				storeValue(job, request, dataHeaderRegexpParameter);
				storeValue(job, request, dataSplitterRegexpParameter);
				storeValue(job, request, trimQuotesParameter);
				storeValue(job, request, ignoreRegexpParameter);
				storeValue(job, request, dataFooterRegexpParameter);
				storeValue(job, request, minDataColumnsParameter);
				storeValue(job, request, maxDataColumnsParameter);
				storeValue(job, request, ri.getParameter(Parameters.CHARSET_PARAMETER));
				storeValue(job, request, ri.getParameter(Parameters.DECIMAL_SEPARATOR_PARAMETER));
				
				// Column mappings
				storeValue(job, request, complexMappings);
				boolean allowComplex = "allow".equals(request.getParameterValue(complexMappings.getName()));
				for (PluginParameter<String> pp : getAllColumnMappings(reporterListContext))
				{
					String mapExpression = (String)request.getParameterValue(pp.getName());
					checkColumnMapping(mapExpression, allowComplex, pp.getLabel());
					storeValue(job, pp, mapExpression);
				}
				
				// Error handling parameters
				storeValue(job, request, defaultErrorParameter);
				storeValue(job, request, stringTooLongErrorParameter);
				storeValue(job, request, invalidUseOfNullErrorParameter);
				storeValue(job, request, numberFormatErrorParameter);
				storeValue(job, request, numberOutOfRangeErrorParameter);
				storeValue(job, request, reporterIsUsedErrorParameter);
				
				File importFile = (File)job.getValue(fileParameter.getName());
				if (reporterListContext)
				{
					ReporterList reporterList = (ReporterList)job.getValue("reporterList");
					response.setSuggestedJobName("Import reporters to list '" + reporterList.getName()
							+ "' from '" + importFile.getName() + "'");
				}
				else
				{
					response.setSuggestedJobName("Import reporters from '" + importFile.getName() + "'");
				}
				response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}
	// -------------------------------------------

	/*
		From the AbstractFlatFileReporter class
		-------------------------------------------
	*/
	private Mapper idMapper;
	private Mapper nameMapper;
	private Mapper symbolMapper;
	private Mapper descriptionMapper;
	private Mapper scoreMapper;
	private Mapper reporterTypeMapper;
	private Map<ExtendedProperty, Mapper> extendedMappers;
	private ReporterType reporterType;
	private Map<String, ReporterType> reporterTypes;
	private ReporterList reporterList;
	private boolean updateExisting;
	private boolean createMissing;
	private boolean deleteMode;
	private int numInserted;
	private int numUpdated;
	private int numExists;
	private int numAddedToList;
	private int numNotFound;
	private int numUsed;
	private int numDeleted;
	private Map<String, Float> deferred;
	private Set<String> deleted;
	private FlatFileParser ffp;
	private boolean nullIfException;
	private boolean failIfUsed;
	
	/**
		Create a {@link DbControl} and a {@link ReporterBatcher}. 
		Load the {@link ReporterType} if one has been specified.
	*/
	@Override
	protected void begin(FlatFileParser ffp)
		throws BaseException
	{
		super.begin(ffp);
		dc = sc.newDbControl();
		batcher = ReporterBatcher.getNew(dc);
		reporterType = (ReporterType)getJobOrConfigurationValue("reporterType");
		reporterList = (ReporterList)job.getValue("reporterList");
		String mode = (String)job.getValue("mode");
		if (mode == null) mode = "update";
		if (reporterList != null) 
		{
			reporterList = ReporterList.getById(dc, reporterList.getId());
			deferred = new HashMap<String, Float>();
			numAddedToList = 0;
			deleteMode = "remove".equals(mode);
			updateExisting = false;
			createMissing = false;
		}
		else
		{
			updateExisting = "update".equals(mode);
			createMissing = updateExisting || "create".equals(mode);
			deleteMode = "delete".equals(mode);
		}
		this.ffp = ffp;
		this.nullIfException = "null".equals(job.getValue("numberFormatError"));
		ffp.setUseNullIfException(nullIfException);
		String reporterIsUsedError = getErrorOption("reporterIsUsedError");
		this.failIfUsed = "fail".equals(reporterIsUsedError);
		if (reporterIsUsedError != null)
		{
			addErrorHandler(ItemInUseException.class, new SimpleErrorHandler(!failIfUsed));
		}
		if (deleteMode)
		{
			// Keep track of already deleted reporters
			deleted = HibernateUtil.getDbEngine().caseInsensitiveComparison() ?
				new TreeSet<String>(String.CASE_INSENSITIVE_ORDER) : new HashSet<String>();
		}
		if (job.getValue("file") != null)
		{
			batcher.setUpdateSource("File: " + ((File)job.getValue("file")).getName());
		}
		numInserted = 0;
		numUpdated = 0;
		numExists = 0;
	}

	/**
		Initialise column <code>Mapper</code>:s.
	*/
	@Override
	protected void beginData()
	{
		boolean cropStrings = ("crop".equals(job.getValue("stringTooLongError")));
		// Mapper that always return null
		Mapper nullMapper = new ConstantMapper((String)null);
		idMapper = getMapper(ffp, (String)getJobOrConfigurationValue("reporterIdColumnMapping"), 
			cropStrings ? ReporterData.MAX_EXTERNAL_ID_LENGTH : null, nullMapper);
		if (!deleteMode)
		{
			nameMapper = getMapper(ffp, (String)getJobOrConfigurationValue("nameColumnMapping"), 
				cropStrings ? ReporterData.MAX_NAME_LENGTH : null, null);
			symbolMapper = getMapper(ffp, (String)getJobOrConfigurationValue("symbolColumnMapping"), 
				cropStrings ? ReporterData.MAX_SYMBOL_LENGTH : null, null);
			descriptionMapper = getMapper(ffp, (String)getJobOrConfigurationValue("descriptionColumnMapping"), 
				cropStrings ? ReporterData.MAX_DESCRIPTION_LENGTH : null, null);
			reporterTypeMapper = getMapper(ffp, (String)getJobOrConfigurationValue("reporterTypeColumnMapping"), 
				cropStrings ? ReporterType.MAX_NAME_LENGTH : null, null);
			scoreMapper = getMapper(ffp, (String)getJobOrConfigurationValue("scoreColumnMapping"), null, null);
			
			reporterTypes = new HashMap<String, ReporterType>();
			extendedMappers = new HashMap<ExtendedProperty, Mapper>();
			List<ExtendedProperty> extendedProperties = ExtendedProperties.getProperties("ReporterData");
			if (extendedProperties != null)
			{
				for (ExtendedProperty ep : extendedProperties)
				{
					String name = "extendedColumnMapping." + ep.getName();
					Mapper m = getMapper(ffp, (String)getJobOrConfigurationValue(name), 
						cropStrings && ep.getLength() > 0 ? ep.getLength() : null, null);
					if (m != null) extendedMappers.put(ep, m);
				}
			}
		}
	}
	
	/**
		Close and commit/rollback the ReporterBatcher and DbControl.
	*/
	@Override
	protected void end(boolean success)
		throws BaseException
	{
		try
		{
			batcher.flush();
			if (reporterList != null)
			{
				// Add newly inserted reporters to the reporter list
				for (Map.Entry<String, Float> entry : deferred.entrySet())
				{
					ReporterData reporter = batcher.getByExternalId(entry.getKey());
					if (reporterList.addReporter(reporter, entry.getValue()))
					{
						numAddedToList++;
					}
				}
			}
			batcher.close();
			if (success)
			{
				dc.commit();
			}
			else
			{
				dc.close();
			}
		}
		catch (BaseException ex)
		{
			dc.close();
			throw ex;
		}
		finally
		{
			super.end(success);
		}
	}

	@Override
	protected void handleData(FlatFileParser.Data data)
		throws BaseException
	{
		String externalId = idMapper.getString(data);
		if (externalId == null) throw new InvalidUseOfNullException("externalId");
		ReporterData reporter =  null;
		
		if (deleteMode)
		{
			if (deleted.add(externalId))
			{
				try
				{
					reporter = batcher.getByExternalId(externalId);
				}
				catch (ItemNotFoundException ex)
				{
					numNotFound++;
				}
				if (reporter != null)
				{
					if (reporterList != null)
					{
						reporterList.removeReporter(reporter);
						numDeleted++;
					}
					else
					{
						try
						{
							batcher.delete(reporter);
							batcher.flushDelete();
							numDeleted++;
						}
						catch (BaseException ex)
						{
							if (failIfUsed) throw new ItemInUseException("Reporter[externalId=" + externalId + "]", ex);
							numUsed++;
						}
					}
				}
			}
		}
		else
		{
			// Is the reporter already queued for insert?
			if (!batcher.isInInsertQueue(externalId))
			{
				// No, it's not... try to load it from the database
				try
				{
					reporter = batcher.getByExternalId(externalId);
				}
				catch (ItemNotFoundException ex)
				{
					// It wasn't in the database either, create a new reporter
					numNotFound++;
					if (createMissing)
					{
						reporter = Reporter.getNew(externalId);
					}
				}
			}
	
			// If we have a reporter object, we must set the properties or add it to a reporter list
			if (reporter != null)
			{
				int currentId = reporter.getId();
				if (reporterList != null)
				{
					// Add to reporter list
					Float score = scoreMapper == null ? 
							reporterList.getScore(reporter) : scoreMapper.getFloat(data);
					if (currentId == 0)
					{
						// It is a new reporter, we must wait to add it until the batcher has been flushed
						deferred.put(externalId, score);
					}
					else
					{
						if (reporterList.addReporter(reporter, score))
						{
							numAddedToList++;
						}
					}
				}
				
				// The actual reporter needs updating or it is a new one
				if (updateExisting || currentId == 0)
				{
					
					if (reporterType != null) 
					{
						Reporter.setReporterType(reporter, reporterType);
					}
					else if (reporterTypeMapper != null)
					{
						Reporter.setReporterType(reporter, getReporterType(reporterTypeMapper.getString(data)));
					}
					String name = nameMapper == null ? reporter.getName() : nameMapper.getString(data);
					reporter.setName(name == null ? externalId : name);
					
					if (symbolMapper != null) reporter.setSymbol(symbolMapper.getString(data));
					if (descriptionMapper != null) reporter.setDescription(descriptionMapper.getString(data));
					for (Map.Entry<ExtendedProperty, Mapper> entry : extendedMappers.entrySet())
					{
						ExtendedProperty ep = entry.getKey();
						Mapper m = entry.getValue();
						reporter.setExtended(ep.getName(), m.getByType(data, ep.getType()));
					}
					if (currentId != 0)
					{
						batcher.update(reporter);
						numUpdated++;
					}
					else
					{
						batcher.insert(reporter);
						numInserted++;
					}
				}
				else
				{
					numExists++;
				}
			}
		}
	}
	/**
		Return <code>x new reporters; y updated reporters; z reporters added to list</code>.
	*/
	@Override
	protected String getSuccessMessage(int skippedLines)
	{
		String msg;
		if (reporterList != null)
		{
			if (deleteMode)
			{
				msg = numDeleted + " reporter(s) removed from list";
			}
			else
			{
				msg = numAddedToList + " reporter(s) added to list";
			}
			if (numNotFound > 0) msg += "; " + numNotFound + " reporter(s) ignored because they didn't exist";
		}
		else
		{
			if (deleteMode)
			{
				msg = numDeleted + " deleted reporter(s)";
				if (numUsed > 0) msg += "; " + numUsed + " reporter(s) are used and could not be deleted";
				if (numNotFound > 0) msg += "; " + numNotFound + " reporter(s) ignored because they didn't exist";
			}
			else
			{
				msg = numInserted + " new reporter(s)";
				if (numUpdated > 0) msg += "; " + numUpdated + " updated reporter(s)";
				if (numExists > 0)
				{
					msg += "; " + numExists + " reporter(s) skipped (already existed)";	
				}
			}
		}
		if (skippedLines > 0) msg += "; " + skippedLines + " line(s) skipped due to errors";
		return msg;
	}
	// -------------------------------------------
	
	private List<PluginParameter<String>> getAllColumnMappings(boolean reporterListContext)
	{
		if (allColumnMappings == null)
		{
			// Column mappings
			allColumnMappings = new ArrayList<PluginParameter<String>>();
			if (reporterListContext)
			{
				allColumnMappings.add(cloneParameterWithDefaultValue(reporterIdColumnMapping));
				allColumnMappings.add(cloneParameterWithDefaultValue(scoreColumnMapping));
			}
			else
			{
				allColumnMappings.add(cloneParameterWithDefaultValue(nameColumnMapping));
				allColumnMappings.add(cloneParameterWithDefaultValue(reporterIdColumnMapping));
				allColumnMappings.add(cloneParameterWithDefaultValue(descriptionColumnMapping));
				allColumnMappings.add(cloneParameterWithDefaultValue(symbolColumnMapping));
				allColumnMappings.add(cloneParameterWithDefaultValue(reporterTypeColumnMapping));
				List<ExtendedProperty> extended = ExtendedProperties.getProperties("ReporterData");
				if (extended != null)
				{
					for (ExtendedProperty ep : extended)
					{
						String name = "extendedColumnMapping."+ep.getName();
						allColumnMappings.add(
							new PluginParameter<String>(
								name,
								ep.getTitle(),
								ep.getDescription(),
								(String)getJobOrConfigurationValue(name),
								optionalColumnMapping
								)
							);
					}
				}
			}
		}
		return allColumnMappings;
	}

	private RequestInformation getConfigureJobParameters(GuiContext context)
	{
		if (configureJob == null)
		{
			// RequestInformation for CONFIGURE_JOB
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
			boolean reporterListContext = context != null && 
			(context.getItem() == Item.REPORTERLIST || context.getItem() == Item.REPORTERSCORE);
			
			if (reporterListContext)
			{
				ItemParameterType<ReporterList> reporterListType = new ItemParameterType<ReporterList>(ReporterList.class, null, true, 1, null);
				reporterListParameter = new PluginParameter<ReporterList>(
					"reporterList",
					"Reporter list",
					"The list to import reporters to",
					reporterListType);
				parameters.add(reporterListParameter);
			}
			parameters.add(fileParameter);
			
			boolean deletePermission = sc.hasPermission(Permission.DELETE, Item.REPORTER);
			boolean createPermission = sc.hasPermission(Permission.CREATE, Item.REPORTER);
			boolean writePermission = sc.hasPermission(Permission.WRITE, Item.REPORTER);
			
			List<String> allowedModes = new ArrayList<String>();
			String modeDescription = "";
			String defaultMode = null;
			if (reporterListContext)
			{
				allowedModes.add("add");
				allowedModes.add("remove");
				defaultMode = "add";
				modeDescription = 
					"Select the operating mode of the plug-in. When adding/removing to/from a " +
					"reporter list, reporters will not be created or updated.\n\n" +
					"add = Add reporters to the reporter list\n" +
					"remove = Remove reporter from the reporter list";
			}
			else 
			{
				modeDescription = "Select the operating mode of the plug-in.\n\n";
				if (createPermission) 
				{
					allowedModes.add("create");
					modeDescription += "create = Only create missing reporters; existing ones are not updated\n";
				}
				if (writePermission) 
				{
					allowedModes.add("update");
					defaultMode = "update";
					modeDescription += "update = Update existing reporters and create missing ones\n";
				}
				if (deletePermission) 
				{
					allowedModes.add("delete");
					modeDescription += "delete = Delete existing reporters\n";
				}
			}
			StringParameterType modeType = new StringParameterType(255, defaultMode, false, 1, 0, 0, allowedModes);
			PluginParameter<String> modeParameter = new PluginParameter<String>(
					"mode",
					"Mode",
					modeDescription,
					modeType
					);
			
			parameters.add(modeParameter);
			
			// Parser regular expressions
			parameters.add(parserSection);
			parameters.add(cloneParameterWithDefaultValue(headerRegexpParameter));
			parameters.add(cloneParameterWithDefaultValue(dataHeaderRegexpParameter));
			parameters.add(cloneParameterWithDefaultValue(dataSplitterRegexpParameter));
			parameters.add(cloneParameterWithDefaultValue(trimQuotesParameter));
			parameters.add(cloneParameterWithDefaultValue(ignoreRegexpParameter));
			parameters.add(cloneParameterWithDefaultValue(dataFooterRegexpParameter));
			parameters.add(cloneParameterWithDefaultValue(minDataColumnsParameter));
			parameters.add(cloneParameterWithDefaultValue(maxDataColumnsParameter));
			parameters.add(Parameters.charsetParameter(null, null, 
					(String)getJobOrConfigurationValue(Parameters.CHARSET_PARAMETER)));
			parameters.add(Parameters.decimalSeparatorParameter(null, null, 
					(String)getJobOrConfigurationValue(Parameters.DECIMAL_SEPARATOR_PARAMETER)));

			// Column mappings
			parameters.add(mappingSection);
			parameters.add(cloneParameterWithDefaultValue(complexMappings));
			parameters.addAll(getAllColumnMappings(reporterListContext));
			
			parameters.add(errorSection);
			parameters.add(defaultErrorParameter);
			parameters.add(stringTooLongErrorParameter);
			parameters.add(invalidUseOfNullErrorParameter);
			parameters.add(numberFormatErrorParameter);
			parameters.add(numberOutOfRangeErrorParameter);
			if (!reporterListContext)
			{
				parameters.add(reporterIsUsedErrorParameter);
			}
			configureJob = new RequestInformation
			(
				Request.COMMAND_CONFIGURE_JOB,
				"Select a file to import reporters from",
				"Here you select which file to import the reporters from, and if " +
				"existing reporters should be updated or not.",
				parameters
			);
		}
		return configureJob;
	}
	
	private RequestInformation getConfigurePluginParameters(GuiContext context)
	{
		if (configurePlugin == null)
		{
			// Load reporter types and initialise the reporterTypeParameter
			DbControl dc = sc.newDbControl();
			ItemQuery<ReporterType> query = ReporterType.getQuery();
			query.order(Orders.asc(Hql.property("name")));
			List<ReporterType> reporterTypes = new ArrayList<ReporterType>(query.list(dc));
			dc.close();
			
			ItemParameterType<ReporterType> reporterTypeType = new ItemParameterType<ReporterType>(ReporterType.class, null, false, 1, reporterTypes);
			reporterTypeParameter = new PluginParameter<ReporterType>(
				"reporterType",
				"Reporter type",
				"The reporter type assigned to the imported reporters",
				reporterTypeType);
					
			// Parameters for CONFIGURE_PLUGIN
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();

			// Reporter type
			parameters.add(reporterTypeParameter);
			
			// Parser regular expressions
			parameters.add(parserSection);
			parameters.add(cloneParameterWithDefaultValue(headerRegexpParameter));
			parameters.add(cloneParameterWithDefaultValue(dataHeaderRegexpParameter));
			parameters.add(cloneParameterWithDefaultValue(dataSplitterRegexpParameter));
			parameters.add(cloneParameterWithDefaultValue(trimQuotesParameter));
			parameters.add(cloneParameterWithDefaultValue(ignoreRegexpParameter));
			parameters.add(cloneParameterWithDefaultValue(dataFooterRegexpParameter));
			parameters.add(cloneParameterWithDefaultValue(minDataColumnsParameter));
			parameters.add(cloneParameterWithDefaultValue(maxDataColumnsParameter));
			parameters.add(Parameters.charsetParameter(null, null, 
					(String)getJobOrConfigurationValue(Parameters.CHARSET_PARAMETER)));
			parameters.add(Parameters.decimalSeparatorParameter(null, null, 
					(String)getJobOrConfigurationValue(Parameters.DECIMAL_SEPARATOR_PARAMETER)));

			// Column mappings
			parameters.add(mappingSection);
			parameters.add(cloneParameterWithDefaultValue(complexMappings));
			parameters.addAll(getAllColumnMappings(false));
			
			configurePlugin = new RequestInformation
			(
				Request.COMMAND_CONFIGURE_PLUGIN,
				"Parser settings",
				"Enter the regular expressions used to parse the text file " +
				"and the column mappings used to find matching properties for the reporters.",
				parameters
			);

		}
		return configurePlugin;
	}
	
	private ReporterType getReporterType(String name)
	{
		if (name == null)
		{
			return null;
		}
		
		ReporterType rt = reporterTypes.get(name);
		if (rt == null)
		{
			DbControl dc = sc.newDbControl();
			try
			{
				ItemQuery<ReporterType> q = ReporterType.getQuery();
				q.restrict(Restrictions.eq(Hql.property("name"), Expressions.parameter("name", name, Type.STRING)));
				List<ReporterType> rtList = q.list(dc);
				if (rtList.size() == 0)
				{
					rt = ReporterType.getNew(dc);
					rt.setName(name);
					dc.saveItem(rt);
					dc.commit();
					reporterTypes.put(name, rt);
				}
				else 
				{
					rt = rtList.get(0);
				}
			}
			finally
			{
				dc.close();
			}
		}
		return rt;
		
	}
}
