/**
	$Id $

	Copyright (C) 2020 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.batchimport;

import java.io.InputStream;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.basedb.core.AnyToAny;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.ItemList;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Metadata;
import net.sf.basedb.core.Nameable;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.Project;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.ParameterValues;
import net.sf.basedb.core.plugin.ParameterValuesWrapper;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.plugins.AbstractFlatFileImporter;
import net.sf.basedb.plugins.batchimport.FileImporter.PathIdMethod;
import net.sf.basedb.plugins.util.Parameters;
import net.sf.basedb.util.Enumeration;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.parser.ConfigureByExample;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;
import net.sf.basedb.util.parser.FlatFileParser.Data;
/**
	Plug-in for importing any-to-any links from simple text files. The plug-in supports
	all files that can be parsed with the {@link FlatFileParser} class. This plug-in
	works without a configuration but can use a configuration to store regular
	expressions and other settings for the flat file parser. In both cases,
	the job configuration is a two-step process:
	
	<ol>
	<li>Setup regular expressions and other options for the flat file parser. If a
		configuration is used all values should already be filled in. In this step
		a file to import from must also be selected.
	<li>Setup error handling options and other settings for the plug-in.
	</ol>

	@author nicklas
	@version 3.17
*/
public class AnyToAnyImporter
	extends AbstractFlatFileImporter
	implements InteractivePlugin, ConfigureByExample
{

	private static volatile Set<GuiContext> guiContexts;

	// second step in the wizard
	private static final String CONFIGURE_OPTIONS = "configure_options";

	private static final StringParameterType optionalColumnMapping = new StringParameterType(255, null, false);
	private static final StringParameterType requiredColumnMapping = new StringParameterType(255, null, true);
	private static final BooleanParameterType optionalBooleanType = new BooleanParameterType(null, false);

	private static final PluginParameter<String> identificationSection = new PluginParameter<String>(
		"idSection",
		"Item identification",
		"Options related to how find existing items.", 
		null
	);

	private static final PluginParameter<String> idColumnMapping = new PluginParameter<String>(
		"idColumnMapping",
		"Source item identifier",
		"Mapping that picks the identifier of the source item. This " +
		"parameter is required and should map to the name, internal id or external id as " +
		"specified by the <b>Identification method</b> options. Example: \\Name\\",
		requiredColumnMapping
	);
	
	private static final PluginParameter<String> linkNameColumnMapping = new PluginParameter<String>(
		"linkNameColumnMapping",
		"Link name",
		"Mapping that picks the name of the link to the target item. This " +
		"parameter is required and can either be specified in the file or " +
		"as a static value. Example: \\Link\\; Link",
		requiredColumnMapping
	);
	
	private static final PluginParameter<String> targetItemTypeColumnMapping = new PluginParameter<String>(
		"targetItemTypeColumnMapping",
		"Target item type",
		"Mapping that picks the type or the target item the link is pointing to. " +
		"This parameter is required in create and update mode and can either be " + 
		"specified in the file or as a static value. Example: \\Type\\; FILE",
		optionalColumnMapping
	);
	
	private static final PluginParameter<String> targetItemIdColumnMapping = new PluginParameter<String>(
		"targetItemIdColumnMapping",
		"Target item identifier",
		"Mapping that picks the name or ID of a target item from the data columns. " +
		"This parameter is required in create and update mode. The plug-in will first " +
		"try to find an item with the given name. If none is found and the value is numeric it will " +
		"try to load by internal ID. " +
		"Example: \\File\\",
		optionalColumnMapping
	);
	
	protected static final PluginParameter<String> linkDescriptionColumnMapping = new PluginParameter<String>(
		"linkDescriptionColumnMapping",
		"Link description",
		"Mapping that picks the description of the link from the data columns." +
		"Example: \\Description\\",
		optionalColumnMapping
		);
	
	private static final PluginParameter<String> itemNotFoundErrorParameter = new PluginParameter<String>(
		"itemNotFoundError",
		"Source item not found",
		"What to do if the plug-in can't find a source item with the name or ID from the file.\n\n"+
		"skip = Skip the current data line and continue\n"+
		"fail = Stop with an error message",
		new StringParameterType(255, null, false, 1, 0, 0, 
			Arrays.asList( new String[] { "skip", "fail"} ))
	);
	
	private static final PluginParameter<String> multipleItemsFoundErrorParameter = new PluginParameter<String>(
		"multipleItemsFoundError",
		"Multiple source items found",
		"What to do if the plug-in finds more than one source item with the same name or ID.\n\n"+
		"skip = Skip the current data line and continue\n"+
		"fail = Stop with an error message",
		new StringParameterType(255, null, false, 1, 0, 0, 
			Arrays.asList( new String[] { "skip", "fail"} ))
	);

	private static final PluginParameter<String> targetNotFoundErrorParameter = new PluginParameter<String>(
		"targetNotFoundError",
		"Target item not found",
		"What to do if the plug-in can't find a target item.\n\n"+
		"skip = Skip the current data line and continue.\n"+
		"fail = Stop with an error message",
		new StringParameterType(255, null, false, 1, 0, 0, 
			Arrays.asList( new String[] { "skip", "fail"} ))
	);

	private static final PluginParameter<String> multipleTargetsFoundErrorParameter = new PluginParameter<String>(
		"multipleTargetsFoundError",
		"Multiple target items found",
		"What to do if the plug-in finds more than one target item with the same name or ID.\n\n"+
		"skip = Skip the current data line and continue.\n"+
		"fail = Stop with an error message",
		new StringParameterType(255, null, false, 1, 0, 0, 
			Arrays.asList( new String[] { "skip", "fail"} ))
	);

	private static final PluginParameter<String> noWritePermissionToItemErrorParameter = new PluginParameter<String>(
		"noWritePermissionToItemError",
		"Not granted write permission to item",
		"What to do if the plug-in finds an existing item but has no write permission to it.\n\n"+
		"skip = Skip the current data line and continue.\n"+
		"fail = Stop with an error message",
		new StringParameterType(255, null, false, 1, 0, 0, 
			Arrays.asList( new String[] { "skip", "fail"} ))
	);
	
	private static final PluginParameter<String> noUsePermissionToTargetErrorParameter = new PluginParameter<String>(
		"noUsePermissionToTargetError",
		"Not granted use permission to target",
		"What to do if the plug-in finds a target but has no permission to use it.\n\n"+
		"skip = Skip the current data line and continue.\n"+
		"fail = Stop with an error message",
		new StringParameterType(255, null, false, 1, 0, 0, 
			Arrays.asList( new String[] { "skip", "fail"} ))
	);

	public AnyToAnyImporter()
	{}
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	@Override
	public boolean requiresConfiguration()
	{
		return false;
	}
	@Override
	public boolean supportsConfigurations()
	{
		return true;
	}
	// -------------------------------------------

	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	/**
		This plug-in works in list context of all {@link Nameable} items,
		except bioassay sets, bioassays and wells because they are not standalone
		items. We use {@link Metadata#getAnnotatableItems()} to create the contexts.
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		if (guiContexts == null)
		{
			synchronized (AnyToAnyImporter.class)
			{
				Set<GuiContext> tmp = new HashSet<GuiContext>();
				for (Item item : Metadata.getNameableItems())
				{
					if (item.getDefinedPermissions() != null)
					{
						tmp.add(new GuiContext(item, GuiContext.Type.LIST));
					}
				}
				guiContexts = Collections.unmodifiableSet(tmp);
			}
		}
		return guiContexts;
	}

	@Override
	public String isInContext(GuiContext context, Object item)
	{
		List<String> itemTypes = configuration.getValues("itemTypes");
		Item itemType = context.getItem();
		if (itemTypes == null || itemTypes.isEmpty() || itemTypes.contains(itemType.name()))
		{
			return null;
		}
		return "Not used for " + itemType.toString().toLowerCase() + " items";
	}	

	@Override
	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
		{
			requestInformation = getConfigureParserParameters(context, false);
		}
		else if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureParserParameters(context, true);
		}
		else if (command.equals(CONFIGURE_OPTIONS))
		{
			requestInformation = getConfigureOptionsParameters(context);
		}
		return requestInformation;
	}

	@Override
	public void configure(GuiContext context, Request request, Response response) 
	{
		String command = request.getCommand();
		DbControl dc = null;
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN) || command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				boolean forJob = command.equals(Request.COMMAND_CONFIGURE_JOB);
				RequestInformation ri = getConfigureParserParameters(context, true);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
					return;
				}

				ParameterValues wrapper = new ParameterValuesWrapper(request, job, configuration, forJob);
				FlatFileParser parser = null;
				if (forJob)
				{
					// Check that file can be parsed with given regular expressions
					parser = getInitializedFlatFileParser(wrapper);
					File f = (File)request.getParameterValue("file");
					InputStream in = f.getDownloadStream(0);
					parser.setInputStream(in, getCharset(request));
					FlatFileParser.LineType result = parser.parseHeaders();
					in.close();
					if (result == FlatFileParser.LineType.UNKNOWN)
					{
						response.setError("The file could not be parsed with the given settings, " +
							"no data found after " + parser.getLineCount() + " lines.", null);
						return;
					}
				}
				else
				{
					storeValues(configuration, request, ri.getParameter("itemTypes"));
				}

				dc = sc.newDbControl();
				
				// Check that a mapping has been provided for the selected idMethod
				IdMethod idMethod = getIdMethod(dc, context == null ? null : context.getItem(), (String)request.getParameterValue("idMethod"));
				if (idMethod == null)
				{
					response.setError("Unknown method for item identification: " + 
							request.getParameterValue("idMethod"), null);
					return;
				}
				
				String mode = request.getParameterValue("mode");
				if (mode == null || mode.contains("create") || mode.contains("update"))
				{
					if (request.getParameterValue("targetItemTypeColumnMapping") == null)
					{
						response.setError("No column mapping provided for target item type", null);
						return;
					}
					if (request.getParameterValue("targetItemIdColumnMapping") == null)
					{
						response.setError("No column mapping provided for target item identifier", null);
						return;
					}
				}

				// Check the mapping expressions
				boolean allowComplex = "allow".equals(request.getParameterValue(complexMappings.getName()));
				for (PluginParameter<?> pp : ri.getParameters())
				{
					if (pp.getName().contains("ColumnMapping"))
					{
						checkColumnMapping(parser, request.getParameterValue(pp.getName()), allowComplex, pp.getLabel());
					}
				}
				
				// Everything is ok, save values
				if (forJob) storeValue(job, request, fileParameter);
				storeValue(wrapper, request, excelSheetParameter);
				storeValue(wrapper, request, ri.getParameter("idMethod"));
				storeValue(wrapper, request, ri.getParameter("mode"));
				
				// Include options
				if (forJob)
				{
					storeValue(job, request, ri.getParameter("memberOfList"));
					storeValue(job, request, ri.getParameter("includeMine"));
					storeValue(job, request, ri.getParameter("includeShared"));
					storeValue(job, request, ri.getParameter("includeInProject"));
					storeValue(job, request, ri.getParameter("includeOthers"));
				}

				storeValue(wrapper, request, dataHeaderRegexpParameter);
				storeValue(wrapper, request, dataSplitterRegexpParameter);
				storeValue(wrapper, request, trimQuotesParameter);
				storeValue(wrapper, request, ignoreRegexpParameter);
				storeValue(wrapper, request, dataFooterRegexpParameter);
				storeValue(wrapper, request, minDataColumnsParameter);
				storeValue(wrapper, request, maxDataColumnsParameter);
				storeValue(wrapper, request, ri.getParameter(Parameters.CHARSET_PARAMETER));
				
				// Column mappings
				storeValue(wrapper, request, complexMappings);
				storeValue(wrapper, request, idColumnMapping);
				storeValue(wrapper, request, linkNameColumnMapping);
				storeValue(wrapper, request, targetItemTypeColumnMapping);
				storeValue(wrapper, request, targetItemIdColumnMapping);
				storeValue(wrapper, request, linkDescriptionColumnMapping);
				
				if (forJob)
				{
					response.setContinue(CONFIGURE_OPTIONS);
				}
				else
				{
					response.setDone("Plug-in configuration completed");
				}
			}
			
			else if (command.equals(CONFIGURE_OPTIONS))
			{
				RequestInformation ri = getConfigureOptionsParameters(context);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
					return;
				}
			
				// Error handling
				response.setDryRun(Boolean.TRUE.equals(request.getParameterValue(Parameters.DRY_RUN_PARAMETER)));
				for (PluginParameter<?> pp : ri.getParameters())
				{
					if (pp.getParameterType() != null)
					{
						storeValue(job, request, ri.getParameter(pp.getName()));
					}
				}

				File file = (File)job.getValue("file");
				String itemType = context.getItem().toString().toLowerCase();
				String jobName = "Import " + itemType + " links from file '"+file.getName()+"'";
				response.setSuggestedJobName(jobName);
				
				// Save context info
				response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);		
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	// -------------------------------------------

	/*
		From the AbstractFlatFileImporter class
		-------------------------------------------
	*/
	private DbControl dc;
	private Project activeProject;
	private FlatFileParser ffp;
	
	private Item itemType;
	private IdMethod idMethod;
	private boolean createMode;
	private boolean updateMode;
	private boolean removeMode;
	
	private Set<Include> includes;
	private ItemQuery<?> itemQuery;
	private ItemList itemList;
	private Map<Item, ItemQuery<?>> targetItemQueries;
	
	private Mapper idMapper;
	private Mapper linkNameMapper;
	private Mapper targetTypeMapper;
	private Mapper targetIdMapper;
	private Mapper descriptionMapper;
	
	// Error handling
	private boolean dryRun;
	private boolean failIfNotFoundSource;
	private boolean failIfMultipleSourceItems;
	private boolean failIfNotFoundTarget;
	private boolean failIfMultipleFoundTargets;
	private boolean failIfNoWritePermissionToSource;
	private boolean failIfNoUsePermissionToTarget;

	// Stats
	private int numCreated;
	private int numUpdated;
	private int numRemoved;
	private int numSkipped;
	
	/**
		Setup error handling and pre-load some of the configuration options.
	*/
	@Override
	protected void begin(FlatFileParser ffp)
	{
		this.ffp = ffp;
		this.dc = sc.newDbControl();
		if (sc.getActiveProjectId() != 0) 
		{
			activeProject = Project.getById(dc, sc.getActiveProjectId());
		}
		this.itemType = job.getJob().getLastGuiContext().getItem();
		this.idMethod = getIdMethod(dc, itemType, job.getValue("idMethod"));
		String mode = job.getValue("mode");
		this.createMode = mode == null || mode.contains("create");
		this.updateMode = mode == null || mode.contains("update");
		this.removeMode = mode != null && mode.contains("remove");
		log("Mode: " + mode);
		
		this.itemList = (ItemList)job.getValue("itemList");
		if (itemList != null) 
		{
			itemList = ItemList.getById(dc, itemList.getId());
			log("Item list: " + itemList.getName());
		}
		
		// Error handling, we don't use the error handling in AbstractFlatFileImporter
		this.dryRun = job.getJob().isDryRun();
		this.failIfNotFoundSource = "fail".equals(getErrorOption("itemNotFoundError"));
		this.failIfMultipleSourceItems = "fail".equals(getErrorOption("multipleItemsFoundError"));
		this.failIfNotFoundTarget = "fail".equals(getErrorOption("targetNotFoundError"));
		this.failIfMultipleFoundTargets = "fail".equals(getErrorOption("multipleReferencesFoundError"));
		this.failIfNoWritePermissionToSource = "fail".equals(getErrorOption("noWritePermissionToItemError"));
		this.failIfNoUsePermissionToTarget = "fail".equals(getErrorOption("noUsePermissionToTargetError"));
	}

	/**
		Setup column mapping. Creates DbControl and query to find items.
	*/
	@Override
	protected void beginData()
	{
		// Column mappers
		this.idMapper = getMapper(ffp, job.getValue("idColumnMapping"), null, null);
		this.linkNameMapper = getMapper(ffp, job.getValue("linkNameColumnMapping"), null, null);
		if (createMode || updateMode)
		{
			this.targetTypeMapper = getMapper(ffp, job.getValue("targetItemTypeColumnMapping"), null, null);
			this.targetIdMapper = getMapper(ffp, job.getValue("targetItemIdColumnMapping"), null, null);
			this.descriptionMapper = getMapper(ffp, job.getValue("linkDescriptionColumnMapping"), null, null);
		}
		
		// Include options
		this.includes = EnumSet.noneOf(Include.class);
		if (Boolean.TRUE.equals(job.getValue("includeMine"))) includes.add(Include.MINE);
		if (Boolean.TRUE.equals(job.getValue("includeShared"))) includes.add(Include.SHARED);
		if (Boolean.TRUE.equals(job.getValue("includeInProject"))) includes.add(Include.IN_PROJECT);
		if (Boolean.TRUE.equals(job.getValue("includeOthers"))) includes.add(Include.OTHERS);
		itemQuery = idMethod.prepareQuery(dc, createQuery(itemType, itemList));
		if (includes.size() > 0) itemQuery.setIncludes(includes);
		targetItemQueries = new HashMap<>();
	}

	
	@Override
	protected void handleData(Data data)
	{
		String sourceId = idMapper.getString(data);
		String linkName = linkNameMapper.getString(data);
		
		List<? extends BasicItem> result = idMethod.find(dc, itemQuery, sourceId);
		if (result.isEmpty())
		{
			// No item with specified name/external ID was found
			if (failIfNotFoundSource)
			{
				throw new ItemNotFoundException(itemType + "[" + idMethod + "=" + sourceId + "]");
			}
			log("Item not found: " + idMethod + "=" + sourceId, data);
			numSkipped++;
			return;
		}
		else if (result.size() > 1)
		{
			if (failIfMultipleSourceItems)
			{
				throw new BaseException("Found " + result.size() + " " + itemType.toString().toLowerCase() + 
					" items with " + idMethod + "=" + sourceId);
			}
			log("Found " + result.size() + " items: " + idMethod + "=" + sourceId, data);
			numSkipped++;
			return;
		}
		
		BasicItem item = result.get(0);
		AnyToAny link = null;
		BasicItem target = null;
		
		try
		{
			link = AnyToAny.getByName(dc, item, linkName);
			if (removeMode)
			{
				if (!failIfNoWritePermissionToSource && !item.hasPermission(Permission.WRITE))
				{
					log("No write permission for item: " + itemType.name() + "[" + sourceId + "]");
					numSkipped++;
					return;
				}
				
				dc.deleteItem(link);
				log("Link deleted: " + itemType.name() + "[" + sourceId + "][" + linkName + "]", data);
				numRemoved++;
			}
			else if (updateMode)
			{
				String targetId = targetIdMapper.getString(data);
				String targetType = targetTypeMapper.getString(data);				
				target = findTargetItem(targetId, targetType, data);
				
				if (target == null) return;

				if (!failIfNoWritePermissionToSource && !item.hasPermission(Permission.WRITE))
				{
					log("No WRITE permission for item: " + itemType.name() + "[" + sourceId + "]");
					numSkipped++;
					return;
				}
				
				if (!failIfNoUsePermissionToTarget && !target.hasPermission(Permission.USE))
				{
					log("No USE permission for target: " + targetType + "["+targetId +"]");
					numSkipped++;
					return;
				}

				link.setTo(target);
				if (descriptionMapper != null)
				{
					link.setDescription(descriptionMapper.getString(data));
				}
				log("Link updated: " + itemType.name() + "[" + sourceId + "][" + linkName + "] -> " + targetType+"["+targetId+"]", data);
				numUpdated++;
			}
			else
			{
				log("Link already exists: " + itemType.name() + "[" + sourceId + "][" + linkName + "]", data);
				numSkipped++;
			}
		}
		catch (ItemNotFoundException ex)
		{
			if (createMode)
			{
				String targetId = targetIdMapper.getString(data);
				String targetType = targetTypeMapper.getString(data);				
				target = findTargetItem(targetId, targetType, data);
				if (target == null) return;
				
				if (!failIfNoWritePermissionToSource && !item.hasPermission(Permission.WRITE))
				{
					log("No WRITE permission for item: " + itemType.name() + "[" + sourceId + "]");
					numSkipped++;
					return;
				}
				
				if (!failIfNoUsePermissionToTarget && !target.hasPermission(Permission.USE))
				{
					log("No USE permission for target: " + targetType + "["+targetId +"]");
					numSkipped++;
					return;
				}

				link = AnyToAny.getNew(dc, item, target, linkName, false);
				if (descriptionMapper != null)
				{
					link.setDescription(descriptionMapper.getString(data));
				}
				dc.saveItem(link);
				log("Link created: " + itemType.name() + "[" + sourceId + "][" + linkName + "] -> " + targetType+"["+targetId+"]", data);
				numCreated++;
			}
			else
			{
				log("Link not found: " + itemType.name() + "[" + sourceId + "][" + linkName + "]", data);
				numSkipped++;
			}
		}
	}
	
	
	private BasicItem findTargetItem(String targetId, String targetType, Data data)
	{
		Item targetItemType = Item.valueOf(targetType);
		
		ItemQuery<?> targetQuery = targetItemQueries.get(targetItemType); 
		IdMethod targetIdMethod = targetItemType == Item.FILE ? PathIdMethod.PATH_OR_ID : FallbackIdMethod.NAME_OR_ID;
		if (targetQuery == null)
		{
			targetQuery = targetItemType.getQuery();
			if (includes.size() > 0) targetQuery.setIncludes(includes);
			targetIdMethod.prepareQuery(dc, targetQuery);
			targetItemQueries.put(targetItemType, targetQuery);
		}
		
		List<? extends BasicItem> result = targetIdMethod.find(dc, targetQuery, targetId);
		if (result.isEmpty())
		{
			// No item with specified name/external ID was found
			if (failIfNotFoundTarget)
			{
				throw new ItemNotFoundException(targetItemType + " [" + targetIdMethod + "=" + targetId + "]");
			}
			log("Target item not found: " + targetIdMethod + "=" + targetId, data);
			numSkipped++;
			return null;
		}
		else if (result.size() > 1)
		{
			if (failIfMultipleFoundTargets)
			{
				throw new BaseException("Found " + result.size() + " " + itemType.toString().toLowerCase() + 
					" with " + targetIdMethod + "=" + targetId);
			}
			log("Found " + result.size() + " items: " + targetIdMethod + "=" + targetId, data);
			numSkipped++;
			return null;
		}
		
		return result.get(0);
	}
	
	@Override
	protected void end(boolean success)
	{
		try
		{
			if (success)
			{
				if (!dryRun) dc.commit();
			}
		}
		catch (BaseException ex)
		{
			success = false;
			throw ex;
		}
		finally
		{
			if (dc != null) dc.close();
			super.end(success);
		}
	}
	@Override
	protected String getSuccessMessage(int skippedLines)
	{
		String msg = "Parsed " + ffp.getParsedLines() + " lines";
		if (removeMode) msg += "; " + numRemoved + " links removed";
		if (createMode) msg += "; " + numCreated + " links created";
		if (updateMode) msg += "; " + numUpdated + " links updated";
		if (numSkipped > 0) msg += "; " + numSkipped + " lines skipped";
		return msg;
	}
	// -------------------------------------------


	private RequestInformation configureParser;
	private RequestInformation configureOptions;

	private RequestInformation getConfigureParserParameters(GuiContext context, boolean forJob)
	{
		if (configureParser == null)
		{
			DbControl dc = null;
			try
			{
				dc = sc.newDbControl();
				List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
	
				if (forJob)
				{
					// The file to import from
					parameters.add(fileParameter);
					parameters.add(excelSheetParameter);
				}
				else
				{
					Enumeration<String, String> items = new Enumeration<String, String>();
					for (GuiContext ctx : getGuiContexts())
					{
						Item i = ctx.getItem();
						ctx.getItem().name();
						items.add(i.name(), i.toString());
					}
					items.sortValues();
					
					PluginParameter<String> itemTypesParameter =
						new PluginParameter<String>("itemTypes", "Item types", 
							"Specify which types of items this file format is used for. If " +
							"not specified it will be used for all types.",
						new StringParameterType(255, null, false, 0, 0, 0, items));
					
					parameters.add(itemTypesParameter);
				}
				
				PluginParameter<String> modeParameter = new PluginParameter<String>(
					"mode", 
					"Mode", 
					"Select an import mode: \n\n" + 
					"<b>create+update</b> = create new links and update existing links\n"+
					"<b>create</b> = create new links only\n"+
					"<b>update</b> = update existing links only\n"+
					"<b>remove</b> = remove existing links only\n",
					new StringParameterType(255, null, forJob, 1, 0, 0, 
						Arrays.asList(null, "create+update", "create", "update", "remove"))
				);
				parameters.add(modeParameter);
				
				if (forJob)
				{
					parameters.add(identificationSection);
				}
				
				Enumeration<String, String> idMethods = new Enumeration<String, String>();
				for (IdMethod idm : getIdMethods(dc, context == null ? null : context.getItem()))
				{
					idMethods.add(idm.getMethod(), idm.getTitle());
				}
				PluginParameter<String> idMethodParameter = new PluginParameter<String>(
					"idMethod",
					"Identification method",
					"The method to use for finding existing items. You can use either the " +
					"internal ID, the name or the external ID. If not specified, a " +
					"method is automatically selected based on the available column mappings. " +
					"If a method is selected, you need to makes sure that the corresponding " +
					"column mapping is defined.",
					new StringParameterType(null, idMethods.getKey(0), true, 1, 0, 0, idMethods)
				);
				parameters.add(idMethodParameter);

				if (forJob)
				{
					ItemList list = getItemListFromContext(dc, context);
					if (list != null)
					{
						parameters.add(new PluginParameter<ItemList>(
							"memberOfList", "Member of list",
							"The plug-in will only update annotations for items if they are members of this list.",
							new ItemParameterType<>(ItemList.class, list, true, 1, Arrays.asList(list))
						));
					}

					ItemContext cc = sc.getCurrentContext(context.getItem());
					Set<Include> include = cc.getInclude();
					parameters.add(new PluginParameter<Boolean>(
							"includeMine",
							"Owned by me",
							"Search items that are owned by yourself.",
							include.contains(Include.MINE) ? Boolean.TRUE : null,
							optionalBooleanType)
						);
					parameters.add(new PluginParameter<Boolean>(
							"includeShared",
							"Shared to me",
							"Search items that have been shared to you by other users.",
							include.contains(Include.SHARED) ? Boolean.TRUE : null,
							optionalBooleanType)
						);
						
					if (sc.getActiveProjectId() != 0)
					{
						parameters.add(new PluginParameter<Boolean>(
								"includeInProject",
								"In current project",
								"Search items that are shared to the current project.",
								include.contains(Include.IN_PROJECT) ? Boolean.TRUE : null,
								optionalBooleanType)
							);
					}
					if (sc.hasPermission(Permission.READ, context.getItem()))
					{
						parameters.add(new PluginParameter<Boolean>(
								"includeOthers",
								"Owned by others",
								"Search items that are owned by other users.",
								include.contains(Include.OTHERS) ? Boolean.TRUE : null,
								optionalBooleanType)
							);
					}
				}
				
				// Parser regular expressions
				parameters.add(parserSection);
				if (!forJob) parameters.add(excelSheetParameter);
				parameters.add(dataHeaderRegexpParameter);
				parameters.add(dataSplitterRegexpParameter);
				parameters.add(trimQuotesParameter);
				parameters.add(ignoreRegexpParameter);
				parameters.add(dataFooterRegexpParameter);
				parameters.add(minDataColumnsParameter);
				parameters.add(maxDataColumnsParameter);
				parameters.add(Parameters.charsetParameter(null, null, getCharset(), true));
				
				// Mappings for Name and External ID
				parameters.add(mappingSection);
				parameters.add(complexMappings);
				
				parameters.add(idColumnMapping);
				parameters.add(linkNameColumnMapping);
				parameters.add(targetItemTypeColumnMapping);
				parameters.add(targetItemIdColumnMapping);
				parameters.add(linkDescriptionColumnMapping);
				
				if (forJob)
				{
					configureParser = new RequestInformation
					(
						Request.COMMAND_CONFIGURE_JOB,
						"Select a file to import annotations from",
						"Here you select which file to import annotations from, and the " +
						"regular expressions and other settings used to parse it. Step 1 of 3",
						parameters
					);
				}
				else
				{
					configureParser = new RequestInformation
					(
						Request.COMMAND_CONFIGURE_PLUGIN,
						"Specify parser regular expressions and settings",
						"This plug-in can be configured with the regular expressions and " +
						"settings used to parse files. Mapping file columns to annotation types must " +
						"be done for each import.",
						parameters
					);				
				}
			}
			finally
			{
				if (dc != null) dc.close();
			}
		}
		return configureParser;
	}
	
	protected RequestInformation getConfigureOptionsParameters(GuiContext context)
	{
		if (configureOptions == null)
		{
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();

			// Error handling options
			parameters.add(errorSection);
			parameters.add(Parameters.logFileParameter(null, null, null));
			parameters.add(Parameters.dryRunParameter(null, null, null));
			parameters.add(defaultErrorParameter);
			parameters.add(itemNotFoundErrorParameter);
			parameters.add(multipleItemsFoundErrorParameter);
			parameters.add(noWritePermissionToItemErrorParameter);
			
			parameters.add(targetNotFoundErrorParameter);
			parameters.add(noUsePermissionToTargetErrorParameter);
			parameters.add(multipleTargetsFoundErrorParameter);
			
			configureOptions = new RequestInformation
			(
				CONFIGURE_OPTIONS,
				"Other settings and error handling options",
				"Specify other settings and error handling options. Step 2 of 2.",
				parameters
			);
		}
		return configureOptions;	
	}

	private List<IdMethod> getIdMethods(DbControl dc, Item item)
	{
		List<IdMethod> methods = new ArrayList<>();
		methods.add(item == Item.FILE ? PathIdMethod.ROOT : PropertyIdMethod.NAME);
		methods.add(InternalIdMethod.INTERNAL_ID);
		if (item == null || hasExternalId(item)) methods.add(PropertyIdMethod.EXTERNAL_ID);
		return methods;
	}
	
	/**
		Check if the current item has an 'externalId' property. We use reflection to
		look for the 'getExternalId' method in the item's data class.
	*/
	private boolean hasExternalId(Item item)
	{
		Class<?> dataClass = item.getDataClass();
		Method method = null;
		try
		{
			method = dataClass.getMethod("getExternalId");
		}
		catch (NoSuchMethodException ex)
		{}
		return method != null;
	}

	/**

	*/
	protected IdMethod getIdMethod(DbControl dc, Item item, String method)
	{
		for (IdMethod im : getIdMethods(dc, item))
		{
			if (im.getMethod().equals(method))
			{
				return im;
			}
		}
		return PropertyIdMethod.NAME;
	}

	/**
		Create a query that return items of the specified type. We use reflection
		to call the static method 'getQuery' on the item's item class, for
		example {@link net.sf.basedb.core.Sample#getQuery()}.
		
		@param itemType The type of items to search
		@return A query that searches items of the specified type.
	*/
	private ItemQuery<?> createQuery(Item itemType, ItemList list)
	{
		return list != null && list.getMemberType() == itemType ? list.getMembers() : itemType.getQuery();
	}

	
	/**
		If the current context is the "Members" tab for an item list we load
		the item list and will later require that all items to be linked
		are members of that list.
	*/
	private ItemList getItemListFromContext(DbControl dc, GuiContext context)
	{
		if (context == null) return null;
		ItemList list = null;
		String subcontext = context.getSubContext();
		if (subcontext != null && subcontext.startsWith("listmembers"))
		{
			// If subcontext is: "listmembers.nnnn" nnn is the id of the item list
			int listId = Values.getInt(subcontext.split("\\.")[1]);
			list = ItemList.getById(dc, listId);
			// Reject lists with different type of member item (should not happen)
			if (list.getMemberType() != context.getItem()) list = null;
		}
		return list;
	}

}
